"""
GUI support classes for the Reflectivity plug-in.
"""

from typing import Callable, Dict, List, Union

import wx.html

from genx.core.custom_logging import iprint
from genx.gui.help import rst_html
from genx.model import Model
from genx.models.lib.base import AltStrEnum
from genx.models.lib.refl_base import ReflBase as ReflBaseNew
from genx.parameters import Parameters
from genx.plugins.utils import ShowQuestionDialog, ShowWarningDialog

from . import reflectivity_images as images
from .custom_dialog import (ComplexObjectValidator, FloatObjectValidator, NoMatchValidTextObjectValidator,
                            ParameterExpressionCombo, TextObjectValidator, ValidateDialog, ValidateFitDialog,
                            ValidateFitNotebookDialog)
from .reflectivity_misc import ReflectivityModule
from .reflectivity_utils import SampleHandler, find_code_segment

_set_func_prefix = "set"


class MyHtmlListBox(wx.html.HtmlListBox):

    def __init__(self, parent, id, size=(-1, -1), style=wx.BORDER_SUNKEN):
        wx.html.HtmlListBox.__init__(self, parent, id, size=size, style=style)
        self.SetItemList(["Starting up..."])

    def SetItemList(self, html_list):
        self.html_items = html_list
        self.SetItemCount(len(html_list))
        # self.RefreshAll()
        self.Refresh()

    def OnGetItem(self, n):
        return self.html_items[n]


class ReflClassHelpDialog(wx.Frame):
    def __init__(self, parent, item: ReflBaseNew):
        wx.Frame.__init__(
            self,
            parent,
            -1,
            f"{item.__class__.__name__} Help",
            style=wx.MINIMIZE_BOX
            | wx.MAXIMIZE_BOX
            | wx.RESIZE_BORDER
            | wx.FRAME_TOOL_WINDOW
            | wx.CAPTION
            | wx.CLOSE_BOX
            | wx.CLIP_CHILDREN
            | wx.STAY_ON_TOP,
        )

        sizer = wx.BoxSizer(wx.VERTICAL)

        self.html_win = wx.html.HtmlWindow(self, -1, style=wx.NO_FULL_REPAINT_ON_RESIZE)
        sizer.Add(self.html_win, 1, flag=wx.EXPAND, border=20)

        self.SetSizer(sizer)

        sizer.Fit(self)
        self.Layout()
        size = parent.GetSize()
        pos = parent.GetPosition()
        gsize = wx.GetApp().GetTopWindow().GetSize()
        gpos = wx.GetApp().GetTopWindow().GetPosition()
        self.SetSize((600, gsize.y))
        self.SetPosition((pos.x + size.x, gpos.y))

        doc = f"{item.__class__.__name__}\n{'='*len(item.__class__.__name__)}\n\n{item.__class__.__doc__}"
        doc = rst_html(doc)
        self.html_win.SetPage(doc)
        self.html_win.SetHTMLBackgroundColour((230, 230, 255, 255))


class ReflClassEditor:
    """
    Class to handle GUI input for parameters to any ReflBase derived class.

    The class is analyzed and suitable entries are created automatically. The parameters use the value
    supplied when instantiated within the GenX model script, thus numbers can be calculated using formulae.
    """

    parent: wx.Panel
    object_name: str
    eval_func: Callable
    grid_parameters: Parameters

    instance: ReflBaseNew
    ignore_attributes: List[str]
    pars: List[str]
    par_vals: dict
    par_validators: dict
    par_editable: Dict[str, int]

    def __init__(self, parent, object_name, eval_func, grid_parameters=None, ignore_attributes=None):
        self.parent = parent
        self.object_name = object_name
        self.eval_func = eval_func
        self.grid_parameters = grid_parameters
        self.ignore_attributes = ignore_attributes or []
        self.help_dialog = None

        self.initialize()
        self.extract_values()
        self.extract_grid_params()

    def initialize(self):
        """
        Set up the necessary attributes for the analysis.
        """
        self.instance = self.eval_func(self.object_name)
        if not isinstance(self.instance, ReflBaseNew):
            raise RuntimeError(f"{self.instance} is not an instance of a ReflBase derived object")
        self.pars = []
        self.par_vals = {}
        self.par_validators = {}
        self.par_editable = {}

    def extract_values(self):
        """
        Collect information from current model needed to build the dialog window.

        Each ModelParamBase object has a ._ca dictionary that is populated during model
        compilation with a string of the parameter values given on class creation. If
        keyword was not supplied the dictionary contains the default value.
        """
        for par_name, value_info in self.instance._parameter_info().items():
            if par_name in self.ignore_attributes:
                continue
            value = self.instance._ca[par_name]
            validator = self.get_validator(value_info)
            if isinstance(value_info.type, type) and issubclass(value_info.type, AltStrEnum):
                # for AltStrEnum evaluate value string, convert to Enum and then get string representation
                value = str(value_info.type(self.eval_func(value)))
            self.pars.append(par_name)
            self.par_vals[par_name] = value
            self.par_validators[par_name] = validator

    def extract_grid_params(self):
        """
        Checks the grid for fit parameters and overwrites the dialog
        values with those.
        Also sets the editable state accordingly.
        """
        if self.grid_parameters is None:
            return
        for par_name, value_info in self.instance._parameter_info().items():
            if par_name in self.ignore_attributes:
                continue
            func_name = self.object_name + "." + _set_func_prefix + par_name.capitalize()
            if value_info.type is not complex:
                grid_value = self.grid_parameters.get_value_by_name(func_name)
                if grid_value is not None:
                    self.par_vals[par_name] = grid_value
                self.par_editable[par_name] = self.grid_parameters.get_fit_state_by_name(func_name)
            else:
                grid_value_real = self.grid_parameters.get_value_by_name(func_name + "real")
                grid_value_imag = self.grid_parameters.get_value_by_name(func_name + "imag")
                if grid_value_imag is not None or grid_value_real is not None:
                    v = self.eval_func(self.par_vals[par_name])
                    if grid_value_real is not None:
                        v = grid_value_real + v.imag * 1.0j
                    if grid_value_imag is not None:
                        v = v.real + grid_value_imag * 1.0j
                    self.par_vals[par_name] = repr(v)
                self.par_editable[par_name] = max(
                    self.grid_parameters.get_fit_state_by_name(func_name + "real"),
                    self.grid_parameters.get_fit_state_by_name(func_name + "imag"),
                )

    def ShowHelp(self, event):
        if self.help_dialog:
            return
        self.help_dialog = ReflClassHelpDialog(self.dlg, self.instance)
        self.help_dialog.Show()

    def ShowDialog(self, title=None):
        if title is None:
            title = f"{self.instance.__class__.__name__} Editor ({self.object_name})"

        groups = self.instance.Groups
        units = getattr(self.instance, "Units", False)

        self.dlg = ValidateFitDialog(
            self.parent,
            self.pars,
            self.par_vals,
            self.par_validators,
            title=title,
            groups=groups,
            units=units,
            editable_pars=self.par_editable,
        )

        if self.instance.__class__.__doc__ != "":
            help_btn = wx.Button(self.dlg, -1, "Show Help")
            help_btn.SetBackgroundColour((230, 230, 255, 255))
            # dlg.border_sizer.InsertSpacer(0, 2)
            sizer = wx.BoxSizer(wx.HORIZONTAL)
            sizer.AddStretchSpacer(1)
            sizer.Add(help_btn, 0, wx.FIXED_MINSIZE, 0)
            self.dlg.border_sizer.Insert(0, sizer, 0, wx.ALIGN_RIGHT)
            self.dlg.Fit()
            self.dlg.Bind(wx.EVT_BUTTON, self.ShowHelp, help_btn)

        res = self.dlg.ShowModal()
        if res == wx.ID_OK:
            vals = self.dlg.GetValues()
            states = self.dlg.GetStates()
            self.update_object(vals, states)
            self.update_grid(vals, states)
        if self.help_dialog:
            self.help_dialog.Destroy()
        self.dlg.Destroy()
        return res == wx.ID_OK

    def update_object(self, vals, states):
        for name, value_info in self.instance._parameter_info().items():
            if name not in self.pars:
                continue
            if not states[name]:
                # only update parameters in the script that are not set in the grid
                orig_type = value_info.type
                if getattr(orig_type, "__origin__", None) is Union:
                    e_value = self.eval_func(vals[name])
                    success = False
                    for subtype in orig_type.__args__:
                        try:
                            setattr(self.instance, name, subtype(e_value))
                        except (ValueError, TypeError):
                            continue
                        else:
                            self.instance._ca[name] = str(vals[name])
                    return
                elif orig_type is str or issubclass(orig_type, AltStrEnum):
                    e_value = vals[name]
                    self.instance._ca[name] = repr(vals[name])
                else:
                    e_value = self.eval_func(vals[name])
                    self.instance._ca[name] = str(vals[name])
                setattr(self.instance, name, orig_type(e_value))

    def update_grid(self, vals, states):
        vspan = self.parent.variable_span
        for name, value_info in self.instance._parameter_info().items():
            if name not in self.pars:
                continue
            if states[name] and self.par_editable[name] != 3 and self.par_editable[name] != states[name]:
                value = self.eval_func(vals[name])
                # if the state of a parameter is changed to fit or set in grid, update the grid line
                if value_info.type is complex:
                    func_name = self.object_name + "." + _set_func_prefix + name.capitalize() + "real"
                    val = value.real
                    minval = min(val * (1 - vspan), val * (1 + vspan))
                    maxval = max(val * (1 - vspan), val * (1 + vspan))
                    self.grid_parameters.set_fit_state_by_name(func_name, val, states[name], minval, maxval)
                    val = value.imag
                    minval = min(val * (1 - vspan), val * (1 + vspan))
                    maxval = max(val * (1 - vspan), val * (1 + vspan))
                    func_name = self.object_name + "." + _set_func_prefix + name.capitalize() + "imag"
                    self.grid_parameters.set_fit_state_by_name(func_name, val, states[name], minval, maxval)
                else:
                    val = value
                    minval = min(val * (1 - vspan), val * (1 + vspan))
                    maxval = max(val * (1 - vspan), val * (1 + vspan))
                    func_name = self.object_name + "." + _set_func_prefix + name.capitalize()
                    self.grid_parameters.set_fit_state_by_name(func_name, value, states[name], minval, maxval)
            elif self.par_editable[name] and not states[name]:
                # remove fit parameter from grid as is changed to set in script
                if value_info.type is complex or complex in getattr(value_info.type, "__args__", ()):
                    func_name = self.object_name + "." + _set_func_prefix + name.capitalize() + "real"
                    self.grid_parameters.set_fit_state_by_name(func_name, 0.0, 0, 0, 0)
                    func_name = self.object_name + "." + _set_func_prefix + name.capitalize() + "imag"
                    self.grid_parameters.set_fit_state_by_name(func_name, 0.0, 0, 0, 0)
                else:
                    func_name = self.object_name + "." + _set_func_prefix + name.capitalize()
                    self.grid_parameters.set_fit_state_by_name(func_name, 0.0, 0, 0, 0)

    def get_validator(self, value_info):
        if isinstance(value_info.type, type) and issubclass(value_info.type, AltStrEnum):
            validator = [i.value for i in value_info.type]
        elif value_info.type is complex or complex in getattr(value_info.type, "__args__", ()):
            validator = ComplexObjectValidator(eval_func=self.eval_func)
        elif value_info.type is bool:
            validator = ["True", "False"]
        else:
            validator = FloatObjectValidator(eval_func=self.eval_func)
        return validator


class SamplePanel(wx.Panel):
    sampleh: SampleHandler
    model: ReflectivityModule

    def __init__(self, parent, plugin, refindexlist=None):
        wx.Panel.__init__(self, parent)
        if refindexlist is None:
            refindexlist = []
        self.refindexlist = refindexlist
        self.plugin = plugin
        self.variable_span = 0.25

        # Colours indicating different states
        # Green wx.Colour(138, 226, 52), ORANGE wx.Colour(245, 121, 0)
        self.fit_colour = (245, 121, 0)
        # Tango Sky blue wx.Colour(52, 101, 164), wx.Colour(114, 159, 207)
        self.const_fit_colour = (114, 159, 207)

        boxver = wx.BoxSizer(wx.HORIZONTAL)
        boxhor = wx.BoxSizer(wx.VERTICAL)
        self.toolbar = wx.ToolBar(self, style=wx.TB_FLAT | wx.TB_HORIZONTAL)
        self.do_toolbar()
        boxhor.Add(self.toolbar, proportion=0, flag=wx.EXPAND)
        boxhor.AddSpacer(2)
        self.listbox = MyHtmlListBox(self, -1, style=wx.BORDER_SUNKEN)
        self.Bind(wx.EVT_LISTBOX_DCLICK, self.lbDoubleClick, self.listbox)
        boxhor.Add(self.listbox, 1, wx.EXPAND)

        boxver.Add(boxhor, 1, wx.EXPAND)

        self.SetSizer(boxver)
        self.toolbar.Realize()
        self.update_callback = lambda event: ""

    def do_toolbar(self):
        dpi_scale_factor = wx.GetApp().dpi_scale_factor
        tb_bmp_size = int(dpi_scale_factor * 20)

        newid = wx.NewId()
        self.toolbar.AddTool(
            newid,
            "Insert Layer",
            bitmap=wx.Bitmap(images.insert_layer.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
            shortHelp="Insert a Layer",
        )
        self.Bind(wx.EVT_TOOL, self.InsertLay, id=newid)

        newid = wx.NewId()
        self.toolbar.AddTool(
            newid,
            "Insert Stack",
            bitmap=wx.Bitmap(images.insert_stack.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
            shortHelp="Insert a Stack",
        )
        self.Bind(wx.EVT_TOOL, self.InsertStack, id=newid)

        newid = wx.NewId()
        self.toolbar.AddTool(
            newid,
            "Delete",
            bitmap=wx.Bitmap(images.delete.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
            shortHelp="Delete item",
        )
        self.Bind(wx.EVT_TOOL, self.DeleteSample, id=newid)

        newid = wx.NewId()
        self.toolbar.AddTool(
            newid,
            "Rename",
            bitmap=wx.Bitmap(images.change_name.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
            shortHelp="Rename",
        )
        self.Bind(wx.EVT_TOOL, self.ChangeName, id=newid)

        newid = wx.NewId()
        self.toolbar.AddTool(
            newid,
            "Move up",
            bitmap=wx.Bitmap(images.move_up.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
            shortHelp="Move item up",
        )
        self.Bind(wx.EVT_TOOL, self.MoveUp, id=newid)

        newid = wx.NewId()
        self.toolbar.AddTool(
            newid,
            "Move down",
            bitmap=wx.Bitmap(images.move_down.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
            shortHelp="Move item down",
        )
        self.Bind(wx.EVT_TOOL, self.MoveDown, id=newid)

        newid = wx.NewId()
        self.toolbar.AddTool(
            newid,
            "Edit Sample",
            bitmap=wx.Bitmap(images.sample.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
            shortHelp="Edit Sample parameters",
        )
        self.Bind(wx.EVT_TOOL, self.EditSampleParameters, id=newid)

        newid = wx.NewId()
        self.toolbar.AddTool(
            newid,
            "Edit Instrument",
            bitmap=wx.Bitmap(images.instrument.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
            shortHelp="Edit Instruments",
        )
        self.Bind(wx.EVT_TOOL, self.EditInstrument, id=newid)

    def SetUpdateCallback(self, func):
        """SetUpdateCallback(self, func) --> None

        Sets the update callback will be called when the sample is updated.
        The call is on the form func(event)
        """
        self.update_callback = func

    def set_sampleh(self, sampleh: SampleHandler):
        self.sampleh = sampleh

    def set_model(self, model: ReflectivityModule):
        self.model = model

    def create_html_decorator(self):
        """
        creates a html decorator function
        :return:
        """
        grid_parameters = self.plugin.GetModel().get_parameters()
        dic_lookup = {}
        for par in grid_parameters.get_names():
            l = par.split(".")
            if len(l) == 2:
                name = l[0]
                par_name = l[1][3:].lower()
                dic_lookup[(name, par_name)] = (
                    grid_parameters.get_value_by_name(par),
                    grid_parameters.get_fit_state_by_name(par),
                )
        fit_color_str = "rgb(%d,%d,%d)" % self.fit_colour
        const_fit_color_str = "rgb(%d,%d,%d)" % self.const_fit_colour

        def decorator(name, str):
            """Decorator to indicate the parameters that are fitted"""
            try:
                start_index = str.index("(") + 1
            except ValueError:
                start_index = 0
            ret_str = str[:start_index]
            for par_str in str[start_index:].split(","):
                par_name = par_str.split("=")[0].strip()
                # par_name normal parameter (real number)
                if (name, par_name) in dic_lookup:
                    val, state = dic_lookup[(name, par_name)]
                    if state == 1:
                        par_str = " <font color=%s><b>%s=%.2e</b></font>," % (fit_color_str, par_name, val)
                    elif state == 2:
                        par_str = " <font color=%s><b>%s=%.2e</b></font>," % (const_fit_color_str, par_name, val)
                # par_name is a complex parameter...
                elif (name, par_name + "real") in dic_lookup or (name, par_name + "imag") in dic_lookup:
                    if (name, par_name + "real") in dic_lookup:
                        val, state = dic_lookup[(name, par_name + "real")]
                        if state == 1:
                            par_str = " <font color=%s><b>%s=(%.2e,</b></font>" % (fit_color_str, par_name, val)
                        elif state == 2:
                            par_str = " <font color=%s><b>%s=(%.2e,</b></font>" % (const_fit_color_str, par_name, val)
                    else:
                        par_str = " <b>%s=??+</b>" % par_name
                    if (name, par_name + "imag") in dic_lookup:
                        val, state = dic_lookup[(name, par_name + "imag")]
                        if state == 1:
                            par_str += " <font color=%s><b>%.2e)</b></font>," % (fit_color_str, val)
                        elif state == 2:
                            par_str += " <font color=%s><b>%.2e)</b></font>," % (const_fit_color_str, val)
                    else:
                        par_str += " <b>??)</b>,"

                else:
                    par_str += ","
                ret_str += par_str
            # Remove trailing ,
            if ret_str[-1] == ",":
                ret_str = ret_str[:-1]
            if str[-1] == ")" and ret_str[-1] != ")":
                ret_str += ")"
            return ret_str

        return decorator

    def Update(self, update_script=True):
        deco = self.create_html_decorator()
        sl = self.sampleh.getStringList(html_encoding=True, html_decorator=deco)
        self.listbox.SetItemList(sl)
        if update_script:
            self.update_callback(None)

    def SetSample(self, sample, names):
        self.sampleh.sample = sample
        self.sampleh.names = names
        self.Update()

    def EditSampleParameters(self, evt):
        """Event handler that creates a dialog box to edit the sample parameters.

        :param evt:
        :return: Nothing
        """
        obj_name = "sample"
        eval_func = self.plugin.GetModel().eval_in_model

        if isinstance(self.sampleh.sample, ReflBaseNew):
            grid_parameters = self.plugin.GetModel().get_parameters()
            editor = ReflClassEditor(
                self, obj_name, eval_func, grid_parameters, ignore_attributes=["Ambient", "Substrate", "Stacks"]
            )
            if editor.ShowDialog():
                self.plugin.parent.paramter_grid.SetParameters(grid_parameters)
                self.sampleh.sample = self.plugin.GetModel().script_module.sample

                self.Update()
            return

        editable, grid_parameters, groups, pars, units, validators, vals = self.ExtractSampleParameters(obj_name)

        dlg = ValidateFitDialog(
            self, pars, vals, validators, title="Sample Editor", groups=groups, units=units, editable_pars=editable
        )

        if dlg.ShowModal() == wx.ID_OK:
            old_vals = vals
            vals = dlg.GetValues()
            # print vals
            states = dlg.GetStates()
            for par in pars:
                if not states[par]:
                    old_type = type(old_vals[par])
                    if isinstance(self.sampleh.sample, ReflBaseNew):
                        self.sampleh.sample._ca[par] = vals[par]
                        setattr(self.sampleh.sample, par, old_type(vals[par]))
                    else:
                        setattr(self.sampleh.sample, par, old_type(vals[par]))
                if editable[par] != states[par]:
                    value = eval_func(vals[par])
                    minval = min(value * (1 - self.variable_span), value * (1 + self.variable_span))
                    maxval = max(value * (1 - self.variable_span), value * (1 + self.variable_span))
                    func_name = obj_name + "." + _set_func_prefix + par.capitalize()
                    grid_parameters.set_fit_state_by_name(func_name, value, states[par], minval, maxval)
                    # Tell the grid to reload the parameters
                    self.plugin.parent.paramter_grid.SetParameters(grid_parameters)

            self.Update()

        dlg.Destroy()

    def ExtractSampleParameters(self, obj_name):
        grid_parameters = self.plugin.GetModel().get_parameters()
        validators = {}
        vals = {}
        pars = []
        items = []
        editable = {}
        try:
            string_choices = self.model.sample_string_choices
        except Exception as e:
            string_choices = {}
        for item in self.model.SampleParameters:
            if item != "Stacks" and item != "Substrate" and item != "Ambient":
                if item in string_choices:
                    validators[item] = string_choices[item]
                else:
                    validators[item] = FloatObjectValidator()
                val = getattr(self.sampleh.sample, item)
                vals[item] = val
                pars.append(item)
                items.append((item, val))
                # Check if the parameter is in the grid and in that case set it as uneditable
                func_name = obj_name + "." + _set_func_prefix + item.capitalize()
                grid_value = grid_parameters.get_value_by_name(func_name)
                editable[item] = grid_parameters.get_fit_state_by_name(func_name)
                if grid_value is not None:
                    vals[item] = grid_value
        try:
            groups = self.model.SampleGroups
        except Exception:
            groups = False
        try:
            units = self.model.SampleUnits
        except Exception:
            units = False
        return editable, grid_parameters, groups, pars, units, validators, vals

    def SetInstrument(self, instruments):
        """SetInstrument(self, instrument) --> None

        Sets the instruments should be a dictionary of instruments with the key being the
        name of the instrument
        """
        self.instruments = instruments

    def EditInstrument(self, evt):
        """Event handler that creates a dialog box to edit the instruments.

        :param evt:
        :return: Nothing
        """
        eval_func = self.plugin.GetModel().eval_in_model
        editable, grid_parameters, model_inst_params, pars, validators, vals, groups, units = (
            self.ExtractInstrumentParams()
        )

        old_insts = []
        for inst_name in self.instruments:
            old_insts.append(inst_name)

        dlg = ValidateFitNotebookDialog(
            self,
            pars,
            vals,
            validators,
            title="Instrument Editor",
            groups=groups,
            units=units,
            fixed_pages=["inst"],
            editable_pars=editable,
        )

        if isinstance(self.sampleh.sample, ReflBaseNew):

            def show_help(event):
                hlp = ReflClassHelpDialog(dlg, self.sampleh.model.Instrument())
                hlp.Show()

            help_btn = wx.Button(dlg, -1, "Show Help")
            help_btn.SetBackgroundColour((230, 230, 255, 255))
            # dlg.border_sizer.InsertSpacer(0, 2)
            sizer = wx.BoxSizer(wx.HORIZONTAL)
            sizer.AddStretchSpacer(1)
            sizer.Add(help_btn, 0, wx.FIXED_MINSIZE, 0)
            dlg.border_sizer.Insert(0, sizer, 0, wx.ALIGN_RIGHT)
            dlg.Fit()
            dlg.Bind(wx.EVT_BUTTON, show_help, help_btn)

        if dlg.ShowModal() == wx.ID_OK:
            old_vals = vals
            vals = dlg.GetValues()
            # print vals
            states = dlg.GetStates()
            self.UpdateInstrumentConfiguration(
                editable, eval_func, grid_parameters, model_inst_params, old_vals, states, vals
            )

            # Loop to remove instrument from grid if not returned from Dialog
            for inst_name in old_insts:
                if inst_name not in list(vals.keys()):
                    for par in pars:
                        if editable[inst_name][par] > 0:
                            func_name = inst_name + "." + _set_func_prefix + par.capitalize()
                            grid_parameters.set_fit_state_by_name(func_name, 0, 0, 0, 0)

            # Tell the grid to reload the parameters
            self.plugin.parent.paramter_grid.SetParameters(grid_parameters)

            for change in dlg.GetChanges():
                if change[0] != "" and change[1] != "":
                    self.plugin.InstrumentNameChange(change[0], change[1])
                elif change[1] == "":
                    self.plugin.InstrumentNameChange(change[0], "inst")

            self.Update()
        else:
            pass
        dlg.Destroy()

    def UpdateInstrumentConfiguration(
        self, editable, eval_func, grid_parameters, model_inst_params, old_vals, states, vals
    ):
        if isinstance(self.sampleh.sample, ReflBaseNew):
            return self.UpdateInstrumentConfigurationNew(
                editable, eval_func, grid_parameters, model_inst_params, old_vals, states, vals
            )
        self.instruments = {}
        for inst_name in vals:
            new_instrument = False
            if inst_name not in self.instruments:
                # A new instrument must be created:
                self.instruments[inst_name] = self.model.Instrument()
                new_instrument = True
            for par in self.model.InstrumentParameters:
                if not states[inst_name][par]:
                    orig_type = type(model_inst_params[par])
                    if orig_type is str:
                        e_value = vals[inst_name][par]
                    else:
                        e_value = eval_func(vals[inst_name][par])
                    setattr(self.instruments[inst_name], par, orig_type(e_value))
                else:
                    setattr(self.instruments[inst_name], par, old_vals[inst_name][par])
                if new_instrument and states[inst_name][par] > 0:
                    value = eval_func(vals[inst_name][par])
                    minval = min(value * (1 - self.variable_span), value * (1 + self.variable_span))
                    maxval = max(value * (1 - self.variable_span), value * (1 + self.variable_span))
                    func_name = inst_name + "." + _set_func_prefix + par.capitalize()
                    grid_parameters.set_fit_state_by_name(func_name, value, states[inst_name][par], minval, maxval)
                elif not new_instrument:
                    if editable[inst_name][par] != states[inst_name][par]:
                        value = eval_func(vals[inst_name][par])
                        minval = min(value * (1 - self.variable_span), value * (1 + self.variable_span))
                        maxval = max(value * (1 - self.variable_span), value * (1 + self.variable_span))
                        func_name = inst_name + "." + _set_func_prefix + par.capitalize()
                        grid_parameters.set_fit_state_by_name(func_name, value, states[inst_name][par], minval, maxval)

    def UpdateInstrumentConfigurationNew(
        self, editable, eval_func, grid_parameters, model_inst_params, old_vals, states, vals
    ):
        self.instruments = {}
        for inst_name in vals:
            new_instrument = False
            if inst_name not in self.instruments:
                # A new instrument must be created:
                self.instruments[inst_name] = self.model.Instrument()
                new_instrument = True
            inst = self.instruments[inst_name]
            for name, value_info in inst._parameter_info().items():
                if not states[inst_name][name]:
                    orig_type = value_info.type
                    if orig_type is str or issubclass(orig_type, AltStrEnum):
                        e_value = vals[inst_name][name]
                    else:
                        e_value = eval_func(vals[inst_name][name])
                    setattr(self.instruments[inst_name], name, orig_type(e_value))
                elif states[inst_name][name] != 3:
                    setattr(self.instruments[inst_name], name, old_vals[inst_name][name])
                if states[inst_name][name] == 3:
                    # ignore attribute overwritten in script and replace by default
                    orig_type = value_info.type
                    e_value = getattr(inst, name)
                    setattr(self.instruments[inst_name], name, orig_type(e_value))
                elif new_instrument and states[inst_name][name] > 0:
                    value = eval_func(vals[inst_name][name])
                    minval = min(value * (1 - self.variable_span), value * (1 + self.variable_span))
                    maxval = max(value * (1 - self.variable_span), value * (1 + self.variable_span))
                    func_name = inst_name + "." + _set_func_prefix + name.capitalize()
                    grid_parameters.set_fit_state_by_name(func_name, value, states[inst_name][name], minval, maxval)
                elif not new_instrument:
                    if editable[inst_name][name] != states[inst_name][name]:
                        value = eval_func(vals[inst_name][name])
                        minval = min(value * (1 - self.variable_span), value * (1 + self.variable_span))
                        maxval = max(value * (1 - self.variable_span), value * (1 + self.variable_span))
                        func_name = inst_name + "." + _set_func_prefix + name.capitalize()
                        grid_parameters.set_fit_state_by_name(func_name, value, states[inst_name][name], minval, maxval)

    def ExtractInstrumentParams(self):
        if isinstance(self.sampleh.sample, ReflBaseNew):
            return self.ExtractInstrumentParamsNew()
        validators = {}
        vals = {}
        editable = {}
        grid_parameters = self.plugin.GetModel().get_parameters()
        for inst_name in self.instruments:
            vals[inst_name] = {}
            editable[inst_name] = {}
        pars = []
        model_inst_params = self.model.InstrumentParameters
        for item in model_inst_params:
            if item in self.model.instrument_string_choices:
                # validators.append(self.model.instrument_string_choices[item])
                validators[item] = self.model.instrument_string_choices[item]
            else:
                # validators.append(FloatObjectValidator())
                validators[item] = FloatObjectValidator()
            for inst_name in self.instruments:
                val = getattr(self.instruments[inst_name], item)
                vals[inst_name][item] = val
                # Check if the parameter is in the grid and in that case set it as uneditable
                func_name = inst_name + "." + _set_func_prefix + item.capitalize()
                grid_value = grid_parameters.get_value_by_name(func_name)
                editable[inst_name][item] = grid_parameters.get_fit_state_by_name(func_name)
                if grid_value is not None:
                    vals[inst_name][item] = grid_value
            pars.append(item)

        try:
            groups = self.model.InstrumentGroups
        except Exception:
            groups = False
        try:
            units = self.model.InstrumentUnits
        except Exception:
            units = False

        return editable, grid_parameters, model_inst_params, pars, validators, vals, groups, units

    def ExtractInstrumentParamsNew(self):
        validators = {}
        vals = {}
        editable = {}
        grid_parameters = self.plugin.GetModel().get_parameters()
        pars = []
        groups = []
        units = []
        for inst_name in self.instruments:
            inst = self.instruments[inst_name]
            vals[inst_name] = {}
            editable[inst_name] = {}
            model_inst_params = []
            groups = inst.Groups
            units = getattr(inst, "Units", [])
            for name, value_info in inst._parameter_info().items():
                val = getattr(self.instruments[inst_name], name)
                val, validator = self.get_validator(val, value_info)
                model_inst_params.append(name)
                vals[inst_name][name] = val
                validators[name] = validator
                # Check if the parameter is in the grid and in that case set it as uneditable
                if val == "set in script":
                    editable[inst_name][name] = 3
                else:
                    func_name = inst_name + "." + _set_func_prefix + name.capitalize()
                    grid_value = grid_parameters.get_value_by_name(func_name)
                    editable[inst_name][name] = grid_parameters.get_fit_state_by_name(func_name)
                    if grid_value is not None:
                        vals[inst_name][name] = grid_value
                    pars.append(name)

        return editable, grid_parameters, model_inst_params, pars, validators, vals, groups, units

    def get_validator(self, val, value_info, eval_func=eval):
        try:
            val = value_info.type(val)
        except Exception:
            pass
        if issubclass(value_info.type, AltStrEnum):
            validator = [i.value for i in value_info.type]
            val = str(val)
        elif not isinstance(val, value_info.type):
            # current value does not correspond to type, should have been converted on class creation
            val = "set in script"
            validator = ["set in script"]
        elif value_info.type is complex or complex in getattr(value_info.type, "__args__", ()):
            validator = ComplexObjectValidator(eval_func=eval_func)
        elif value_info.type is bool:
            validator = ["True", "False"]
            val = repr(val)
        else:
            validator = FloatObjectValidator(eval_func=eval_func)
        return val, validator

    def MoveUp(self, evt):
        sl = self.sampleh.moveUp(self.listbox.GetSelection())
        if sl:
            self.Update()
            self.listbox.SetSelection(self.listbox.GetSelection() - 1)

    def MoveDown(self, evt):
        sl = self.sampleh.moveDown(self.listbox.GetSelection())
        if sl:
            self.Update()
            self.listbox.SetSelection(self.listbox.GetSelection() + 1)

    def InsertStack(self, evt):
        # Create Dialog box
        validators = {"Name": NoMatchValidTextObjectValidator(self.sampleh.names)}
        vals = {}
        pars = ["Name"]
        vals["Name"] = "name"
        dlg = ValidateDialog(self, pars, vals, validators, title="Give Stack Name")

        # Show the dialog
        if dlg.ShowModal() == wx.ID_OK:
            vals = dlg.GetValues()
        dlg.Destroy()
        # if not a value is selected operate on first
        pos = max(self.listbox.GetSelection(), 0)
        sl = self.sampleh.insertItem(pos, "Stack", vals["Name"])
        if sl:
            self.Update()
        else:
            self.plugin.ShowWarningDialog("Can not insert a stack at the" " current position.")

    def InsertLay(self, evt):
        # Create Dialog box
        # items = [('Name', 'name')]
        # validators = [NoMatchValidTextObjectValidator(self.sampleh.names)]
        dlg = ValidateDialog(
            self,
            ["Name"],
            {"Name": "name"},
            {"Name": NoMatchValidTextObjectValidator(self.sampleh.names)},
            title="Give Layer Name",
        )
        # Show the dialog
        if dlg.ShowModal() == wx.ID_OK:
            vals = dlg.GetValues()
        else:
            vals = {"Name": "name"}
        dlg.Destroy()
        # if not a value is selected operate on first
        pos = max(self.listbox.GetSelection(), 0)
        # Create the Layer
        sl = self.sampleh.insertItem(pos, "Layer", vals["Name"])
        if sl:
            self.Update()
        else:
            self.plugin.ShowWarningDialog(
                "Can not insert a layer at the" " current position. Layers has to be part of a stack."
            )

    def DeleteSample(self, evt):
        slold = self.sampleh.getStringList()
        sl = self.sampleh.deleteItem(self.listbox.GetSelection())
        if sl:
            self.Update()

    def ChangeName(self, evt):
        """Change the name of the current selected item."""
        pos = self.listbox.GetSelection()
        if pos == 0 or pos == len(self.sampleh.names) - 1:
            self.plugin.ShowInfoDialog(
                "It is forbidden to change the" "name of the substrate (Sub) and the Ambient (Amb) layers."
            )
        else:
            unallowed_names = self.sampleh.names[:pos] + self.sampleh.names[max(0, pos - 1) :]
            dlg = ValidateDialog(
                self,
                ["Name"],
                {"Name": self.sampleh.names[pos]},
                {"Name": NoMatchValidTextObjectValidator(unallowed_names)},
                title="Give New Name",
            )

            if dlg.ShowModal() == wx.ID_OK:
                vals = dlg.GetValues()
                result = self.sampleh.changeName(pos, vals["Name"])
                if result:
                    self.Update()
                else:
                    iprint("Unexpected problems when changing name...")
            dlg.Destroy()

    def lbDoubleClick(self, evt):
        sel = self.sampleh.getItem(self.listbox.GetSelection())
        if isinstance(sel, self.model.Layer):
            sl = self.EditLayer(sel)
        else:
            sl = self.EditStack(sel)

        if sl:
            self.Update()

    def EditStack(self, sel):
        if isinstance(sel, ReflBaseNew):
            return self.EditStackNew()
        obj_name = self.sampleh.getName(self.listbox.GetSelection())
        eval_func = self.plugin.GetModel().eval_in_model
        sl = None
        items = []
        validators = {}
        vals = {}
        pars = []
        editable = {}
        grid_parameters = self.plugin.GetModel().get_parameters()
        # The selected item is a Stack
        if isinstance(sel, ReflBaseNew):
            for name, value_info in sel._parameter_info().items():
                if name != "Layers":
                    value = getattr(sel, name)
                    if value_info.type in [int, float]:
                        validators[name] = FloatObjectValidator(eval_func, alt_types=[self.model.Stack])
                    else:
                        validators[name] = TextObjectValidator()
                    items.append((name, value))
                    pars.append(name)
                    vals[name] = value

                    # Check if the parameter is in the grid and in that case set it as uneditable
                    func_name = obj_name + "." + _set_func_prefix + name.capitalize()
                    grid_value = grid_parameters.get_value_by_name(func_name)
                    editable[name] = grid_parameters.get_fit_state_by_name(func_name)
                    if grid_value is not None:
                        vals[name] = grid_value

            try:
                groups = self.model.StackGroups
            except Exception:
                groups = False
            try:
                units = self.model.StackUnits
            except Exception:
                units = False
        else:
            for item in list(self.model.StackParameters.keys()):
                if item != "Layers":
                    value = getattr(sel, item)
                    if isinstance(value, float):
                        validators[item] = FloatObjectValidator(eval_func, alt_types=[self.model.Stack])
                    else:
                        validators[item] = TextObjectValidator()
                    items.append((item, value))
                    pars.append(item)
                    vals[item] = value

                    # Check if the parameter is in the grid and in that case set it as uneditable
                    func_name = obj_name + "." + _set_func_prefix + item.capitalize()
                    grid_value = grid_parameters.get_value_by_name(func_name)
                    editable[item] = grid_parameters.get_fit_state_by_name(func_name)
                    if grid_value is not None:
                        vals[item] = grid_value

            try:
                groups = self.model.StackGroups
            except Exception:
                groups = False
            try:
                units = self.model.StackUnits
            except Exception:
                units = False
        dlg = ValidateFitDialog(
            self, pars, vals, validators, title="Layer Editor", groups=groups, units=units, editable_pars=editable
        )
        if dlg.ShowModal() == wx.ID_OK:
            vals = dlg.GetValues()
            states = dlg.GetStates()
            for par in pars:
                if not states[par]:
                    setattr(sel, par, vals[par])
                if editable[par] != states[par]:
                    value = eval_func(vals[par])
                    minval = min(value * (1 - self.variable_span), value * (1 + self.variable_span))
                    maxval = max(value * (1 - self.variable_span), value * (1 + self.variable_span))
                    func_name = obj_name + "." + _set_func_prefix + par.capitalize()
                    grid_parameters.set_fit_state_by_name(func_name, value, states[par], minval, maxval)
                    # Does not seem to be necessary
                    self.plugin.parent.paramter_grid.SetParameters(grid_parameters)
            sl = self.sampleh.getStringList()
        dlg.Destroy()
        return sl

    def EditStackNew(self):
        obj_name = self.sampleh.getName(self.listbox.GetSelection())
        eval_func = self.plugin.GetModel().eval_in_model
        sl = None
        grid_parameters = self.plugin.GetModel().get_parameters()
        editor = ReflClassEditor(self, obj_name, eval_func, grid_parameters, ignore_attributes=["Layers"])
        if editor.ShowDialog():
            self.plugin.parent.paramter_grid.SetParameters(grid_parameters)
            self.sampleh.sample = self.plugin.GetModel().script_module.sample
            sl = self.sampleh.getStringListNew()
        return sl

    def EditLayer(self, sel):
        if isinstance(sel, ReflBaseNew):
            return self.EditLayerNew()
        obj_name = self.sampleh.getName(self.listbox.GetSelection())
        eval_func = self.plugin.GetModel().eval_in_model
        sl = None
        items = []
        validators = {}
        vals = {}
        pars = []
        editable = {}
        grid_parameters = self.plugin.GetModel().get_parameters()
        # The selected item is a Layer
        for item in self.model.LayerParameters.keys():
            value = getattr(sel, item)
            vals[item] = value
            # if item!='n' and item!='fb':
            if type(self.model.LayerParameters[item]) != type(1 + 1.0j):
                # Handle real parameters
                validators[item] = FloatObjectValidator(eval_func, alt_types=[self.model.Layer])
                func_name = obj_name + "." + _set_func_prefix + item.capitalize()
                grid_value = grid_parameters.get_value_by_name(func_name)
                if grid_value is not None:
                    vals[item] = grid_value
                editable[item] = grid_parameters.get_fit_state_by_name(func_name)

            else:
                # Handle complex parameters
                validators[item] = ComplexObjectValidator(eval_func, alt_types=[self.model.Layer])
                func_name = obj_name + "." + _set_func_prefix + item.capitalize()
                grid_value_real = grid_parameters.get_value_by_name(func_name + "real")
                grid_value_imag = grid_parameters.get_value_by_name(func_name + "imag")
                if grid_value_real is not None:
                    v = eval_func(vals[item]) if type(vals[item]) is str else vals[item]
                    vals[item] = grid_value_real + v.imag * 1.0j
                if grid_value_imag is not None:
                    v = eval_func(vals[item]) if type(vals[item]) is str else vals[item]
                    vals[item] = v.real + grid_value_imag * 1.0j
                editable[item] = max(
                    grid_parameters.get_fit_state_by_name(func_name + "real"),
                    grid_parameters.get_fit_state_by_name(func_name + "imag"),
                )

            items.append((item, value))
            pars.append(item)

            # Check if the parameter is in the grid and in that case set it as uneditable
            # func_name = obj_name + '.' + _set_func_prefix + item.capitalize()
            # grid_value = grid_parameters.get_value_by_name(func_name)
            # editable[item] = grid_parameters.get_fit_state_by_name(func_name)

        try:
            groups = self.model.LayerGroups
        except Exception:
            groups = False
        try:
            units = self.model.LayerUnits
        except Exception:
            units = False
        dlg = ValidateFitDialog(
            self, pars, vals, validators, title="Layer Editor", groups=groups, units=units, editable_pars=editable
        )
        if dlg.ShowModal() == wx.ID_OK:
            vals = dlg.GetValues()
            states = dlg.GetStates()
            for par in pars:
                if not states[par]:
                    setattr(sel, par, vals[par])
                if editable[par] != states[par]:
                    value = eval_func(vals[par])

                    if type(value) is complex:
                        # print type(value)
                        func_name = obj_name + "." + _set_func_prefix + par.capitalize() + "real"
                        val = value.real
                        minval = min(val * (1 - self.variable_span), val * (1 + self.variable_span))
                        maxval = max(val * (1 - self.variable_span), val * (1 + self.variable_span))
                        grid_parameters.set_fit_state_by_name(func_name, val, states[par], minval, maxval)
                        val = value.imag
                        minval = min(val * (1 - self.variable_span), val * (1 + self.variable_span))
                        maxval = max(val * (1 - self.variable_span), val * (1 + self.variable_span))
                        func_name = obj_name + "." + _set_func_prefix + par.capitalize() + "imag"
                        grid_parameters.set_fit_state_by_name(func_name, val, states[par], minval, maxval)
                    else:
                        val = value
                        minval = min(val * (1 - self.variable_span), val * (1 + self.variable_span))
                        maxval = max(val * (1 - self.variable_span), val * (1 + self.variable_span))
                        func_name = obj_name + "." + _set_func_prefix + par.capitalize()
                        grid_parameters.set_fit_state_by_name(func_name, value, states[par], minval, maxval)

                    # Does not seem to be necessary
                    self.plugin.parent.paramter_grid.SetParameters(grid_parameters)
            sl = self.sampleh.getStringList()
        dlg.Destroy()
        return sl

    def EditLayerNew(self):
        obj_name = self.sampleh.getName(self.listbox.GetSelection())
        eval_func = self.plugin.GetModel().eval_in_model
        sl = None
        grid_parameters = self.plugin.GetModel().get_parameters()
        editor = ReflClassEditor(self, obj_name, eval_func, grid_parameters)
        if editor.ShowDialog():
            self.plugin.parent.paramter_grid.SetParameters(grid_parameters)
            self.sampleh.sample = self.plugin.GetModel().script_module.sample
            sl = self.sampleh.getStringListNew()
        return sl


class DataParameterPanel(wx.Panel):
    """Widget that defines parameters coupling and different parameters
    for different data sets.
    """

    def __init__(self, parent, plugin):
        wx.Panel.__init__(self, parent)
        self.plugin = plugin
        boxver = wx.BoxSizer(wx.VERTICAL)
        # Indention for a command - used to separate commands and data
        self.command_indent = "<pre>   "
        self.script_update_func = None
        self.parameterlist = []

        self.toolbar = wx.ToolBar(self, style=wx.TB_FLAT | wx.TB_HORIZONTAL)
        self.do_toolbar()
        boxver.Add(self.toolbar, proportion=0, flag=wx.EXPAND, border=1)
        boxver.AddSpacer(4)

        self.listbox = MyHtmlListBox(self, -1, style=wx.BORDER_SUNKEN)
        self.Bind(wx.EVT_LISTBOX_DCLICK, self.Edit, self.listbox)
        boxver.Add(self.listbox, 1, wx.EXPAND)

        self.SetSizer(boxver)

    def do_toolbar(self):
        dpi_scale_factor = wx.GetApp().dpi_scale_factor
        tb_bmp_size = int(dpi_scale_factor * 20)

        button_names = ["Insert", "Delete", "User Variables"]
        button_images = [
            wx.Bitmap(images.add.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
            wx.Bitmap(images.delete.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
            wx.Bitmap(images.custom_parameters.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
        ]
        callbacks = [self.Insert, self.Delete, self.EditPars]
        tooltips = ["Insert a command", "Delete command", "Edit user variables"]

        for i in range(len(button_names)):
            newid = wx.NewId()
            self.toolbar.AddTool(newid, label=button_names[i], bitmap=button_images[i], shortHelp=tooltips[i])
            self.Bind(wx.EVT_TOOL, callbacks[i], id=newid)
        self.toolbar.Realize()

    def onsimulate(self, event):
        """Function to simulate the model.

        :return:
        """
        self.plugin.parent.eh_tb_simulate(event)

    def SetDataList(self, datalist):
        """SetDataList(self, datalist) --> None

        Sets the name of the available data sets
        """
        self.datalist = datalist

    def GetDataList(self):
        """SetDataList(self) --> list

        Retrives the data list
        """
        return self.datalist

    def SetParameterList(self, parameterlist):
        """SetParameterList(self, parameterlist) --> None

        Sets the code list for all definition of custom parameters
        """
        self.parameterlist = parameterlist

    def GetParameterList(self):
        """SetParameterList(self) --> list

        Retrives the parameter list
        """
        return self.parameterlist

    def SetExpressionList(self, expressionlist):
        """SetExpressionList(expressionlist) --> None

        Sets the expression list, should be a 2D list with the
        one list for each item in datalist
        """
        if len(expressionlist) != len(self.datalist):
            raise ValueError("The list of expression has to have the" + " same length as the data list")
        self.expressionlist = expressionlist

    def GetExpressionList(self):
        """GetExpressionList(self) --> expression list

        Returns the expressionlist
        """
        return self.expressionlist

    def SetSimArgs(self, sim_funcs, insts, args):
        """SetSimArgs(self, sim_func, inst, args) --> None

        Sets the current simulation function for each data set
        their instruments and arguments
        sim_func: A list of simulation names
        inst: A list of instrument names
        args: A list of argument list for the sim_func's each argument should be a string.
        """
        if len(sim_funcs) != len(self.datalist):
            raise ValueError("The list of sim_funcs has to have the" + " same length as the data list")
        if len(insts) != len(self.datalist):
            raise ValueError("The list of insts has to have the" + " same length as the data list")
        if len(args) != len(self.datalist):
            raise ValueError("The list of args has to have the" + " same length as the data list")
        self.sim_funcs = sim_funcs[:]
        self.insts = insts[:]
        self.args = args[:]

    def GetSimArgs(self):
        """GetSimArgs(self) --> (sim_funcs, insts, args)

        See SetSimArgs for a description of the parameters
        """
        return self.sim_funcs, self.insts, self.args

    def AppendSim(self, sim_func, inst, args):
        """AppendSim(self, sim_func, inst, args) --> None

        Appends a simultion to the Panel lists
        """
        self.sim_funcs.append(sim_func)
        self.insts.append(inst)
        self.args.append(args)

    def InstrumentNameChange(self, old_name, new_name):
        """OnInstrumentNameChange --> None

        Exchanges old_name to new name in the simulations."""

        for i in range(len(self.insts)):
            if self.insts[i] == old_name:
                self.insts[i] = new_name
        self.update_listbox()

    def SetUpdateScriptFunc(self, func):
        """SetUpdateScriptFunc(self, func) --> None

        Sets the function to be called when the script needs to be updated.
        will only be called as func(event)
        """
        self.script_update_func = func

    def UpdateListbox(self, update_script=True):
        """Update(self) --> None

        Update the listbox and runs the callback script_update_func
        """
        self.update_listbox()
        if self.script_update_func and update_script:
            self.script_update_func(None)
        self.Refresh()

    def update_listbox(self):
        """update_listbox(self) --> None

        updates the listbox.
        """
        list_strings = []
        for i in range(len(self.datalist)):
            str_arg = ", ".join(self.args[i])
            list_strings.append(
                "<code><b>%s</b>: %s(%s, %s)</code>"
                " \n" % (self.datalist[i], self.sim_funcs[i], str_arg, self.insts[i])
            )
            for item in self.expressionlist[i]:
                list_strings.append(self.command_indent + "%s</pre>" % item)

        self.listbox.SetItemList(list_strings)

    def get_expression_position(self):
        """get_expression_position(self) --> (dataitem, expression)

        Finds the position in the expression list for a certain item.
        return -1 if it can not be found.
        """
        index = self.listbox.GetSelection()

        if index == wx.NOT_FOUND:
            return -1, -1

        dataindex = -1
        itemindex = -1
        listindex = -1
        for i in range(len(self.datalist)):
            dataindex += 1
            listindex += 1
            itemindex = -1
            if listindex >= index:
                return dataindex, itemindex
            for item in self.expressionlist[i]:
                listindex += 1
                itemindex += 1
                if listindex >= index:
                    return dataindex, itemindex

        # If all other things fail...
        return -1, -1

    def Edit(self, event):
        """Edit(self, event) --> None

        Edits an entry in the list
        """
        data_pos, exp_pos = self.get_expression_position()
        if exp_pos != -1 and data_pos != -1:
            # Editing the expressions for variables
            list_item = self.expressionlist[data_pos][exp_pos]
            dlg = ParameterExpressionDialog(self, self.plugin.GetModel(), list_item, sim_func=self.onsimulate)
            if dlg.ShowModal() == wx.ID_OK:
                exp = dlg.GetExpression()
                self.expressionlist[data_pos][exp_pos] = exp
                self.UpdateListbox()
        if exp_pos == -1 and data_pos != -1:
            # Editing the simulation function and its arguments
            dlg = SimulationExpressionDialog(
                self,
                self.plugin.GetModel(),
                self.plugin.sample_widget.instruments,
                self.sim_funcs[data_pos],
                self.args[data_pos],
                self.insts[data_pos],
                data_pos,
            )
            if dlg.ShowModal() == wx.ID_OK:
                self.args[data_pos] = dlg.GetExpressions()
                self.insts[data_pos] = dlg.GetInstrument()
                self.sim_funcs[data_pos] = dlg.GetSim()
                self.UpdateListbox()

    def Insert(self, event):
        """Insert(self, event) --> None

        Inserts a new operations
        """
        data_pos, exp_pos = self.get_expression_position()
        if data_pos != -1:
            dlg = ParameterExpressionDialog(self, self.plugin.GetModel(), sim_func=self.onsimulate)
            if dlg.ShowModal() == wx.ID_OK:
                exp = dlg.GetExpression()
                if exp_pos == -1:
                    self.expressionlist[data_pos].insert(0, exp)
                else:
                    self.expressionlist[data_pos].insert(exp_pos, exp)
                self.UpdateListbox()

    def Delete(self, event):
        """Delete(self, event) --> None

        Deletes an operation
        """
        data_pos, exp_pos = self.get_expression_position()
        if exp_pos != -1 and data_pos != -1:
            self.expressionlist[data_pos].pop(exp_pos)
            self.UpdateListbox()

    def MoveUp(self, event):
        """MoveUp(self, event) --> None

        Move an operation up
        """
        pass

    def MoveDown(self, event):
        """MoveDown(self, event) --> None

        Moves an operation down
        """
        pass

    def EditPars(self, event):
        """EditPars(self, event) --> None

        Creates a new parameter
        """
        dlg = EditCustomParameters(self, self.plugin.GetModel(), self.parameterlist)
        if dlg.ShowModal() == wx.ID_OK:
            self.parameterlist = dlg.GetLines()
            self.UpdateListbox()
        dlg.Destroy()

    def OnDataChanged(self, event):
        """OnDataChanged(self, event) --> None

        Updated the data list
        """
        self.UpdateListbox(update_script=False)


class EditCustomParameters(wx.Dialog):
    model: Model

    def __init__(self, parent, model, lines):
        wx.Dialog.__init__(self, parent, -1, "Custom parameter editor")
        self.SetAutoLayout(True)
        self.model = model
        self.lines = lines
        self.var_name = "cp"

        sizer = wx.BoxSizer(wx.VERTICAL)
        name_ctrl_sizer = wx.GridBagSizer(2, 3)

        col_labels = ["Name", "Value", "Sigma (for systematic error)"]

        for item, index in zip(col_labels, list(range(len(col_labels)))):
            label = wx.StaticText(self, -1, item)
            name_ctrl_sizer.Add(label, (0, index), flag=wx.ALIGN_LEFT, border=5)

        self.name_ctrl = wx.TextCtrl(self, -1, size=(120, -1))
        name_ctrl_sizer.Add(self.name_ctrl, (1, 0), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
        self.value_ctrl = wx.TextCtrl(self, -1, size=(120, -1))
        name_ctrl_sizer.Add(self.value_ctrl, (1, 1), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
        self.error_ctrl = wx.TextCtrl(self, -1, size=(120, -1))
        name_ctrl_sizer.Add(self.error_ctrl, (1, 2), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
        self.add_button = wx.Button(self, id=wx.ID_ANY, label="Add")
        name_ctrl_sizer.Add(self.add_button, (1, 3), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
        sizer.Add(name_ctrl_sizer)
        self.Bind(wx.EVT_BUTTON, self.OnAdd, self.add_button)

        line = wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW | wx.RIGHT | wx.TOP, 5)

        self.listbox = MyHtmlListBox(self, -1, size=(-1, 150), style=wx.BORDER_SUNKEN)
        self.listbox.SetItemList(self.lines)
        sizer.Add(self.listbox, 1, wx.GROW | wx.ALL, 10)

        self.delete_button = wx.Button(self, id=wx.ID_ANY, label="Delete")
        sizer.Add(self.delete_button, 0, wx.CENTRE, 0)
        self.Bind(wx.EVT_BUTTON, self.OnDelete, self.delete_button)

        button_sizer = wx.StdDialogButtonSizer()
        okay_button = wx.Button(self, id=wx.ID_OK)
        # okay_button.SetDefault()
        button_sizer.AddButton(okay_button)
        button_sizer.AddButton(wx.Button(self, id=wx.ID_CANCEL))
        button_sizer.Realize()
        self.Bind(wx.EVT_BUTTON, self.OnApply, okay_button)

        line = wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW | wx.RIGHT | wx.TOP, 5)

        sizer.Add(button_sizer, 0, wx.ALIGN_RIGHT, 5)
        self.SetSizer(sizer)
        sizer.Fit(self)
        self.Layout()

    def OnApply(self, event):
        """OnApply(self, event) --> None

        Callback for ok button click or apply button
        """
        event.Skip()

    def OnAdd(self, event):
        """OnAdd(self, event) --> None

        Callback for adding an entry
        """
        sigma = self.error_ctrl.GetValue()
        if sigma.strip() == "":
            line = "%s.new_var('%s', %s)" % (self.var_name, self.name_ctrl.GetValue(), self.value_ctrl.GetValue())
        else:
            line = "%s.new_sys_err('%s', %s, %s)" % (
                self.var_name,
                self.name_ctrl.GetValue(),
                self.value_ctrl.GetValue(),
                sigma,
            )
        try:
            self.model.eval_in_model(line)
        except Exception as e:
            result = "Could not evaluate the expression. The python error" + "is: \n" + e.__repr__()
            ShowWarningDialog(self, result, "Error in expression")
        else:
            self.lines.append(line)
            self.listbox.SetItemList(self.lines)

    def OnDelete(self, event):
        """OnDelete(self, event) --> None

        Callback for deleting an entry
        """
        result = "Do you want to delete the expression?\n" + "Remember to check if parameter is used elsewhere!"
        result = ShowQuestionDialog(self, result, "Delete expression?")
        if result:
            self.lines.pop(self.listbox.GetSelection())
            self.listbox.SetItemList(self.lines)

    def GetLines(self):
        """GetLines(self) --> uservars lines [list]

        Returns the list user variables.
        """
        return self.lines


class SimulationExpressionDialog(wx.Dialog):
    """A dialog to edit the Simulation expression"""

    model: Model

    def __init__(self, parent, model, instruments, sim_func, arguments, inst_name, data_index):
        """Creates a SimualtionExpressionDialog.

        model - a Model object.
        instruments - a dictionary of possible instruments
        arguments - the arguments to the simulation function, a list of strings.
        sim_func - a string of the simulation function name.
        inst_name - the name of the current instrument.
        data_index - an integer for the current data index.
        """
        if not model.compiled:
            model.compile_script()

        self.model = model
        self.instruments = instruments
        self.available_sim_funcs = list(self.model.eval_in_model("model.SimulationFunctions.keys()"))
        self.data_index = data_index

        # Do the layout of the dialog
        wx.Dialog.__init__(self, parent, -1, "Simulation editor")
        self.SetAutoLayout(True)

        # Find out the maximum number of arguments to the available sim_funcs
        max_val = -1
        self.sim_args = {}
        self.sim_defaults = {}
        for func in self.available_sim_funcs:
            doc = self.model.eval_in_model("model.SimulationFunctions" '["%s"].__doc__' % func)
            doc_lines = find_code_segment(doc, "Parameters").splitlines()
            max_val = max(len(doc_lines), max_val)
            args = []
            defaults = []
            for line in doc_lines:
                items = line.lstrip().rstrip().split(" ")
                args.append(items[0])
                defaults.append(items[1].replace("data", "d"))
            self.sim_args[func] = args
            self.sim_defaults[func] = defaults

        expressions = {"Instrument": inst_name}
        for arg_name, arg in zip(self.sim_args[sim_func], arguments):
            expressions[arg_name] = arg

        if max_val < 0:
            raise ValueError("Wrongly formatted function docs for the simulation functions")

        gbs = wx.GridBagSizer(2, max_val)

        # Creating the column labels
        col_labels = ["Simulation", "Instrument"]
        [col_labels.append(arg) for arg in self.sim_args[sim_func] if not arg in col_labels]
        self.labels = []
        self.arg_controls = []
        for index in range(2 + max_val):
            label = wx.StaticText(self, -1, "")
            gbs.Add(label, (0, index), flag=wx.ALIGN_LEFT, border=5)
            self.labels.append(label)
            # If the expression is not an instrument or simulation function
            if index > 1:
                exp_ctrl = wx.TextCtrl(self, -1, size=(100, -1))
                gbs.Add(exp_ctrl, (1, index), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
                self.arg_controls.append(exp_ctrl)

        for item, label in zip(col_labels[:2], self.labels[:2]):
            label.SetLabel(item)
        # Creating the text boxes for the arguments
        # Setting the text in the column labels and text controls
        for item, label, arg_ctrl in zip(col_labels[2:], self.labels[2:], self.arg_controls):
            label.SetLabel(item)
            arg_ctrl.SetValue(expressions[item])
            arg_ctrl.SetEditable(True)

        for i in range(len(col_labels) - 2, len(self.arg_controls)):
            self.arg_controls[i].SetEditable(False)
            # self.arg_controls[i].Show(False)
            # self.arg_controls[i].SetValue('NA')

        # Creating the controls
        # Simulation choice control
        self.sim_choice = wx.Choice(self, -1, choices=self.available_sim_funcs)
        self.Bind(wx.EVT_CHOICE, self.on_sim_change, self.sim_choice)
        self.sim_choice.SetSelection(self.available_sim_funcs.index(sim_func))
        gbs.Add(self.sim_choice, (1, 0), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)

        # Instrument choice control
        self.inst_choice = wx.Choice(self, -1, choices=list(self.instruments.keys()))
        # self.Bind(wx.EVT_CHOICE, self.on_inst_change, self.inst_choice)
        self.inst_choice.SetSelection(list(self.instruments.keys()).index(expressions["Instrument"]))
        gbs.Add(self.inst_choice, (1, 1), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)

        button_sizer = wx.StdDialogButtonSizer()
        okay_button = wx.Button(self, wx.ID_OK)
        okay_button.SetDefault()
        button_sizer.AddButton(okay_button)
        button_sizer.AddButton(wx.Button(self, wx.ID_CANCEL))

        button_sizer.Realize()
        self.Bind(wx.EVT_BUTTON, self.on_ok_button, okay_button)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(gbs, 1, wx.GROW | wx.ALL, 10)
        line = wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW | wx.RIGHT | wx.TOP, 5)
        sizer.AddSpacer(5)

        sizer.Add(button_sizer, 0, wx.ALIGN_RIGHT, 5)
        sizer.AddSpacer(5)
        self.SetSizer(sizer)
        sizer.Fit(self)
        self.Layout()

    def on_sim_change(self, evt):
        """Callback for changing the choice widget for the different simulations."""
        new_sim = self.sim_choice.GetStringSelection()
        # Update the column labels
        new_labels = []
        for label, arg_name in zip(self.labels[2:], self.sim_args[new_sim]):
            new_labels.append(label.GetLabel() != arg_name)
            label.SetLabel(arg_name)
        # Clear the remaining column labels
        for label in self.labels[len(self.sim_args[new_sim]) + 2 :]:
            label.SetLabel("")

        # Update the text controls - if needed
        for i in range(len(self.sim_args[new_sim])):
            # if new_labels[i]:
            if True:
                self.arg_controls[i].SetValue(self.sim_defaults[new_sim][i])
            self.arg_controls[i].SetEditable(True)
            # self.arg_controls[i].Show(True)
        # Hide and clear the remaining text controls
        for ctrl in self.arg_controls[len(self.sim_args[new_sim]) :]:
            ctrl.SetEditable(False)
            ctrl.SetValue("")

            # ctrl.Show(False)

    def on_ok_button(self, event):
        """Callback for pressing the ok button in the dialog"""
        expressions = self.GetExpressions()
        # Hack to get it working with d = data[0]
        exec("d = data[%d]" % self.data_index, self.model.script_module.__dict__)
        for exp in expressions:
            try:
                self.model.eval_in_model(exp)
            except Exception as e:
                result = "Could not evaluate expression:\n%s.\n" % exp + " The python error is: \n" + e.__repr__()
                ShowWarningDialog(self, result, "Error in expression")
            else:
                event.Skip()

    def GetExpressions(self):
        """Returns the current expressions in the dialog box"""
        return [ctrl.GetValue() for ctrl in self.arg_controls if ctrl.IsEditable()]

    def GetInstrument(self):
        """Returns the selected instrument, a string"""
        return self.inst_choice.GetStringSelection()

    def GetSim(self):
        """Returns the selected simulation, a string"""
        return self.sim_choice.GetStringSelection()


class ParameterExpressionDialog(wx.Dialog):
    """A dialog for setting parameters for fitting"""

    model: Model

    def __init__(self, parent, model, expression=None, sim_func=None):
        wx.Dialog.__init__(self, parent, -1, "Parameter editor")
        self.SetAutoLayout(True)
        self.model = model
        self.sim_func = sim_func
        if not model.compiled:
            model.compile_script()

        gbs = wx.GridBagSizer(2, 3)

        col_labels = ["Object", "Parameter", "Expression"]

        for item, index in zip(col_labels, list(range(len(col_labels)))):
            label = wx.StaticText(self, -1, item)
            gbs.Add(label, (0, index), flag=wx.ALIGN_LEFT, border=5)

        # Get the objects that should be in the choice boxes
        par_dict = model.get_possible_parameters()
        objlist = []
        funclist = []
        for cl in par_dict:
            obj_dict = par_dict[cl]
            for obj in obj_dict:
                objlist.append(obj)
                funclist.append(obj_dict[obj])

        self.objlist = objlist
        self.funclist = funclist
        self.obj_choice = wx.Choice(self, -1, choices=objlist)
        self.Bind(wx.EVT_CHOICE, self.on_obj_change, self.obj_choice)

        self.func_choice = wx.Choice(self, -1)
        # This will init func_choice
        self.obj_choice.SetSelection(0)

        gbs.Add(self.obj_choice, (1, 0), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
        gbs.Add(self.func_choice, (1, 1), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)

        exp_right = ""
        if expression:
            p = expression.find("(")
            exp_left = expression[:p]
            obj = exp_left.split(".")[0]
            func = exp_left.split(".")[1]
            exp_right = expression[p + 1 : -1]
            obj_pos = [i for i in range(len(objlist)) if objlist[i] == obj]
            if len(obj_pos) > 0:
                self.obj_choice.SetSelection(obj_pos[0])
                self.on_obj_change(None)
                func_pos = [i for i in range(len(funclist[obj_pos[0]])) if funclist[obj_pos[0]][i] == func]
                if len(func_pos) > 0:
                    self.func_choice.SetSelection(func_pos[0])
                else:
                    raise ValueError("The function %s for object %s does not exist" % (func, obj))
            else:
                raise ValueError("The object %s does not exist" % obj)

        # self.expression_ctrl = wx.TextCtrl(self, -1, exp_right,\
        #                       size=(300, -1))

        self.expression_ctrl = ParameterExpressionCombo(par_dict, sim_func, self, -1, exp_right, size=(300, -1))
        gbs.Add(self.expression_ctrl, (1, 2), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)

        button_sizer = wx.StdDialogButtonSizer()
        okay_button = wx.Button(self, wx.ID_OK)
        okay_button.SetDefault()
        button_sizer.AddButton(okay_button)
        button_sizer.AddButton(wx.Button(self, wx.ID_CANCEL))
        # apply_button = wx.Button(self, wx.ID_APPLY)
        # apply_button.SetDefault()
        # button_sizer.AddButton(apply_button)
        button_sizer.Realize()
        self.Bind(wx.EVT_BUTTON, self.OnApply, okay_button)
        # self.Bind(wx.EVT_BUTTON, self.OnApply, apply_button)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(gbs, 1, wx.GROW | wx.ALL, 10)
        line = wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW | wx.RIGHT | wx.TOP, 5)
        sizer.AddSpacer(5)

        sizer.Add(button_sizer, 0, wx.ALIGN_RIGHT, 5)
        sizer.AddSpacer(5)
        self.SetSizer(sizer)
        sizer.Fit(self)
        self.Layout()

    def on_obj_change(self, event):
        """on_obj_change(self, event) --> None

        On changing the object the funclist should be updated
        """
        index = self.obj_choice.GetSelection()
        self.func_choice.SetItems(self.funclist[index])

    def OnApply(self, event):
        """OnApply(self, event) --> None"""
        evalstring = self.GetExpression()
        try:
            self.model.eval_in_model(evalstring)
        except Exception as e:
            result = "Could not evaluate the expression. The python" + "is: \n" + e.__repr__()
            ShowWarningDialog(self, result, "Error in expression")
        else:
            event.Skip()

    def GetExpression(self):
        """GetExpression(self) --> expression

        Yields the string that has been edited in the dialog
        """
        objstr = self.obj_choice.GetStringSelection()
        funcstr = self.func_choice.GetStringSelection()
        set_expression = self.expression_ctrl.GetValue()
        evalstring = "%s.%s(%s)" % (objstr, funcstr, set_expression)

        return evalstring

class ORSOExportDialogHook(wx.FileDialogCustomizeHook):
    export_compact = False

    def __init__(self):
        super().__init__()

    def AddCustomControls(self, customizer):
        self.checkbox = customizer.AddCheckBox("Compact view without explicit stacks")
        self.checkbox.SetValue(self.export_compact)

    def TransferDataFromCustomControls(self):
        self.export_compact = self.checkbox.GetValue()



"""
Testing framework for GenX models. Allows the test case to discover all implemented model
tests and run them in a coherent way.
"""

import unittest
import numpy as np

ModelTestCases = []

class ModelTestCase(unittest.TestCase):
    def __init_subclass__(cls, /, **kwargs):
        super().__init_subclass__(**kwargs)
        ModelTestCases.append(cls)

    @property
    def tth(self):
        # x-points for tth in reflectivity
        return np.linspace(0.01, 6.0, 15)

    @property
    def qz(self):
        # x-points for q in reflectivity
        return np.linspace(0.005, 0.3, 15)



"""
==================================================
:mod:`amor` Amor at SINQ reflectometer data loader
==================================================

Loads the default datafile format used for extracted reflectivity at
the SINQ reflectometer Amor.
"""

import os

import numpy as np

from ..data_loader_framework import Template
from ..utils import ShowWarningDialog


class Plugin(Template):
    wildcard = "*.Rqz"

    def __init__(self, parent):
        Template.__init__(self, parent)
        self.x_col = 0
        self.y_col = 1
        self.e_col = 3
        self.xe_col = 2
        self.comment = "#"
        self.skip_rows = 0
        self.delimiter = None

    def LoadData(self, dataset, filename, data_id=0):
        """LoadData(self, data_item_number, filename) --> none

        Loads the data from filename into the data_item_number.
        """
        # get scan number and polarization channel from filename
        name = os.path.basename(filename)[:-4]
        fhandle = open(filename, "r")
        fline = fhandle.readline()
        while not "q_z" in fline:
            if "Input file indices:" in fline:
                name += fline.split(":")[1].strip()
            if "Extracted states:" in fline:
                name += " (%s)" % (fline.split(":")[1].strip().split(" ")[0])
            fline = fhandle.readline()
            if not fline:
                ShowWarningDialog(self.parent, "Could not load the file: " + filename + " \nWrong format.\n")
                return

        try:
            load_array = np.loadtxt(fhandle, delimiter=self.delimiter, comments=self.comment, skiprows=self.skip_rows)
        except Exception as e:
            ShowWarningDialog(
                self.parent,
                "Could not load the file: "
                + filename
                + " \nPlease check the format.\n\n numpy.loadtxt"
                + " gave the following error:\n"
                + str(e),
            )
            return
        else:
            # For the freak case of only one data point
            if len(load_array.shape) < 2:
                load_array = np.array([load_array])
            # Check so we have enough columns
            if load_array.shape[1] - 1 < max(self.x_col, self.y_col, self.e_col):
                ShowWarningDialog(
                    self.parent,
                    "The data file does not contain"
                    + "enough number of columns. It has "
                    + str(load_array[1])
                    + " columns. Rember that the column index start at zero!",
                )
                # Okay now we have showed a dialog lets bail out ...
                return
            # The data is set by the default Template.__init__ function, neat hu
            # Know the loaded data goes into *_raw so that they are not
            # changed by the transforms
            dataset.x_raw = load_array[:, self.x_col]
            dataset.y_raw = load_array[:, self.y_col]
            dataset.error_raw = load_array[:, self.e_col]
            dataset.set_extra_data("res", load_array[:, self.xe_col], "res")
            # Name the dataset accordign to file name
            dataset.name = name
            # Run the commands on the data - this also sets the x,y, error memebers
            # of that data item.
            dataset.run_command()

            # insert metadata into ORSO compatible fields
            dataset.meta["data_source"]["facility"] = "SINQ@PSI"
            dataset.meta["data_source"]["experiment"]["instrument"] = "Amor"
            dataset.meta["data_source"]["experiment"]["probe"] = "neutron"
            dataset.meta["data_source"]["measurement"]["scheme"] = "angle- and energy-dispersive"
            dataset.meta["reduction"] = {"software": {"name": "eos/amor-reducer"}}



"""
Load all example files, compile the script and make a simple simulation.
"""

import os
import unittest

from glob import glob

from genx import api

BASE_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "genx", "examples")


class TestExamples(unittest.TestCase):
    def test_all(self):
        fls = glob(os.path.join(BASE_DIR, "*.hgx"))
        for i, fl in enumerate(fls):
            with self.subTest(f"Datafile {fl}", i=i):
                model, optimizer = api.load(fl)
                model.simulate()



"""
Test of the models.lib.base classes used in model builds (new style).
"""

import unittest

from dataclasses import dataclass, field

from genx.models.lib.base import AltStrEnum, ModelParamBase


class TestModelParamBase(unittest.TestCase):

    def test_conversion(self):
        @dataclass
        class TestParam(ModelParamBase):
            p_str: str = field(default_factory=str)
            p_int: int = 0
            p_float: float = 0.0

        t = TestParam(p_int=1.1, p_float="1", p_str=13)
        self.assertEqual(t.p_int, 1)
        self.assertEqual(t.p_float, 1.0)
        self.assertEqual(t.p_str, "13")
        self.assertEqual(type(t.p_int), int)
        self.assertEqual(type(t.p_float), float)
        self.assertEqual(type(t.p_str), str)

    def test_extraction(self):
        @dataclass
        class TestParam(ModelParamBase):
            p_str: str = field(default_factory=str)
            p_int: int = 0
            p_float: float = 0.0

        txt = 'TestParam(p_int = 1+1, p_float = 0.01+0.02, p_str="abc"+"def")'
        t: TestParam = eval(txt)
        self.assertEqual(t.p_int, 2)
        self.assertEqual(t.p_float, 0.03)
        self.assertEqual(t.p_str, "abcdef")
        t._extract_callpars(txt)
        self.assertEqual(t._ca["p_int"], "1+1")
        self.assertEqual(t._ca["p_float"], "0.01+0.02")
        self.assertEqual(eval(t._ca["p_str"]), "abcdef")

        txt = "TestParam()"
        t: TestParam = eval(txt)
        self.assertEqual(t.p_int, 0)
        self.assertEqual(t.p_float, 0.0)
        self.assertEqual(t.p_str, "")
        t._extract_callpars(txt)
        self.assertEqual(t._ca["p_int"], "0")
        self.assertEqual(t._ca["p_float"], "0.0")
        self.assertEqual(eval(t._ca["p_str"]), "")

    def test_getter_setter(self):
        @dataclass
        class TestParam(ModelParamBase):
            p_str: str = field(default_factory=str)
            p_int: int = 0
            p_float: float = 0.0
            p_complex: complex = 0 + 0j

        t = TestParam()
        self.assertTrue(hasattr(t, "setP_str"))
        self.assertTrue(hasattr(t, "getP_str"))
        self.assertTrue(hasattr(t, "setP_int"))
        self.assertTrue(hasattr(t, "getP_int"))
        self.assertTrue(hasattr(t, "setP_float"))
        self.assertTrue(hasattr(t, "getP_float"))
        self.assertTrue(hasattr(t, "setP_complexreal"))
        self.assertTrue(hasattr(t, "getP_complexreal"))
        self.assertTrue(hasattr(t, "setP_compleximag"))
        self.assertTrue(hasattr(t, "getP_compleximag"))

        self.assertEqual(t.p_int, t.getP_int())
        t.setP_int(24)
        self.assertEqual(t.p_int, 24)
        self.assertEqual(t.p_complex, t.getP_complexreal() + 1j * t.getP_compleximag())
        t.setP_complexreal(0.1)
        t.setP_compleximag(0.4)
        self.assertEqual(t.p_complex, 0.1 + 0.4j)

    def test_recreate(self):
        @dataclass
        class TestParam(ModelParamBase):
            p_str: str = field(default_factory=str)
            p_int: int = 0
            p_float: float = 0.0

        txt = 'TestParam(p_int = 1+1, p_float = 0.01+0.02, p_str="abc"+"def")'
        t: TestParam = eval(txt)
        # repr without extracted params, build from values
        self.assertEqual(t, eval(t._repr_call()))
        t._extract_callpars(txt)
        # repr with extracted params, should be re-evaluated
        self.assertEqual(t, eval(t._repr_call()))


class TestAltStrEnum(unittest.TestCase):

    def test_equivalency(self):
        class TestAltEnum(AltStrEnum):
            peter = "peter"
            kurt = "kurt"
            alternate1_peter = "gert"
            alternate2_peter = "hans"

        self.assertEqual(TestAltEnum("peter"), TestAltEnum("peter"))
        self.assertEqual(TestAltEnum("peter"), TestAltEnum("gert"))
        self.assertEqual(TestAltEnum("gert"), TestAltEnum("peter"))
        self.assertTrue(TestAltEnum("peter") != TestAltEnum("kurt"))
        self.assertTrue(TestAltEnum("gert") != TestAltEnum("kurt"))

    def test_basic_methods(self):
        class TestAltEnum(AltStrEnum):
            peter = "peter"
            kurt = "kurt"
            alternate1_peter = "gert"
            alternate2_peter = "hans"

        t = TestAltEnum("peter")
        self.assertEqual(str(t), "peter")
        self.assertEqual(eval(repr(t)), "peter")
        d = {t: "peter"}

    def test_indexing(self):
        class TestAltEnum(AltStrEnum):
            peter = "peter"
            kurt = "kurt"
            alternate1_peter = "gert"
            alternate2_peter = "hans"

        self.assertEqual(TestAltEnum("peter"), TestAltEnum(0))
        self.assertEqual(TestAltEnum("kurt"), TestAltEnum(1))
        self.assertEqual(TestAltEnum("gert"), TestAltEnum(2))
        self.assertEqual(TestAltEnum("hans"), TestAltEnum(3))

        with self.assertRaises(ValueError):
            TestAltEnum(4)
        with self.assertRaises(ValueError):
            TestAltEnum("nose")
        with self.assertRaises(ValueError):
            TestAltEnum(-1)



import cmath
import math

import numba

from numpy import complex128, empty, empty_like, pi


@numba.jit(
    numba.float64[:](numba.float64[:], numba.float64, numba.complex128[:], numba.float64[:], numba.float64[:]),
    nopython=True,
    parallel=True,
    cache=True,
)
def ReflNB(theta, lamda, n, d, sigma):
    layers = d.shape[0]
    angles = theta.shape[0]
    n0 = n[-1]

    k = 2.0 * pi / lamda
    R = empty_like(theta)

    pre1 = 2.0 * n0 * k
    pre2 = pi / 180.0

    for ai in numba.prange(angles):
        Qi = pre1 * cmath.sqrt((n[0] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)

        Qj = pre1 * cmath.sqrt((n[1] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
        # Fresnel reflectivity for the interfaces
        rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[0] ** 2)
        Aj = rpj
        Qi = Qj

        for lj in range(2, layers):
            Qj = pre1 * cmath.sqrt((n[lj] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
            # Fresnel reflectivity for the interfaces
            rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[lj - 1] ** 2)

            pj = cmath.exp(1.0j * d[lj - 1] * Qi)
            part = Aj * pj
            Aj = (rpj + part) / (1.0 + part * rpj)

            Qi = Qj
        R[ai] = abs(Aj) ** 2
    return R


@numba.jit(
    numba.complex128[:](numba.float64[:], numba.float64, numba.complex128[:], numba.float64[:], numba.float64[:]),
    nopython=True,
    parallel=True,
    cache=True,
)
def AmpNB(theta, lamda, n, d, sigma):
    layers = d.shape[0]
    angles = theta.shape[0]
    n0 = n[-1]

    k = 2.0 * pi / lamda
    A = empty(theta.shape, dtype=complex128)

    pre1 = 2.0 * n0 * k
    pre2 = pi / 180.0

    for ai in numba.prange(angles):
        Qi = pre1 * cmath.sqrt((n[0] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)

        Qj = pre1 * cmath.sqrt((n[1] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
        # Fresnel reflectivity for the interfaces
        rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[0] ** 2)
        Aj = rpj
        Qi = Qj

        for lj in range(2, layers):
            Qj = pre1 * cmath.sqrt((n[lj] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
            # Fresnel reflectivity for the interfaces
            rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[lj - 1] ** 2)

            pj = cmath.exp(1.0j * d[lj - 1] * Qi)
            part = Aj * pj
            Aj = (rpj + part) / (1.0 + part * rpj)

            Qi = Qj
        A[ai] = Aj
    return A


def Refl(theta, lamda, n, d, sigma, return_int=True):
    if return_int:
        return ReflNB(theta, lamda, n, d, sigma)
    else:
        return AmpNB(theta, lamda, n, d, sigma)


@numba.jit(
    numba.float64[:](numba.float64[:], numba.float64, numba.complex128[:], numba.float64[:], numba.float64[:]),
    nopython=True,
    parallel=True,
    cache=True,
)
def ReflQNB(Q, lamda, n, d, sigma):
    layers = d.shape[0]
    points = Q.shape[0]
    n0 = n[-1]

    Q0 = 4.0 * pi / lamda
    R = empty_like(Q)

    for qi in numba.prange(points):
        Qi = cmath.sqrt((n[0] ** 2 - n0**2) * Q0**2 + n0**2 * Q[qi] ** 2)

        Qj = cmath.sqrt((n[1] ** 2 - n0**2) * Q0**2 + n0**2 * Q[qi] ** 2)
        # Fresnel reflectivity for the interfaces
        rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[0] ** 2)
        Aj = rpj
        Qi = Qj

        for lj in range(2, layers):
            Qj = cmath.sqrt((n[lj] ** 2 - n0**2) * Q0**2 + n0**2 * Q[qi] ** 2)
            # Fresnel reflectivity for the interfaces
            rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[lj - 1] ** 2)

            pj = cmath.exp(1.0j * d[lj - 1] * Qi)
            part = Aj * pj
            Aj = (rpj + part) / (1.0 + part * rpj)

            Qi = Qj
        R[qi] = abs(Aj) ** 2
    return R


@numba.jit(
    numba.complex128[:](numba.float64[:], numba.float64, numba.complex128[:], numba.float64[:], numba.float64[:]),
    nopython=True,
    parallel=True,
    cache=True,
)
def AmpQNB(Q, lamda, n, d, sigma):
    layers = d.shape[0]
    points = Q.shape[0]
    n0 = n[-1]

    Q0 = 4.0 * pi / lamda
    A = empty(Q.shape, dtype=complex128)

    for qi in numba.prange(points):
        Qi = cmath.sqrt((n[0] ** 2 - n0**2) * Q0**2 + n0**2 * Q[qi] ** 2)

        Qj = cmath.sqrt((n[1] ** 2 - n0**2) * Q0**2 + n0**2 * Q[qi] ** 2)
        # Fresnel reflectivity for the interfaces
        rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[0] ** 2)
        Aj = rpj
        Qi = Qj

        for lj in range(2, layers):
            Qj = cmath.sqrt((n[lj] ** 2 - n0**2) * Q0**2 + n0**2 * Q[qi] ** 2)
            # Fresnel reflectivity for the interfaces
            rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[lj - 1] ** 2)

            pj = cmath.exp(1.0j * d[lj - 1] * Qi)
            part = Aj * pj
            Aj = (rpj + part) / (1.0 + part * rpj)

            Qi = Qj
        A[qi] = Aj
    return A


def ReflQ(Q, lamda, n, d, sigma, return_int=True):
    if return_int:
        return ReflQNB(Q, lamda, n, d, sigma)
    else:
        return AmpQNB(Q, lamda, n, d, sigma)


@numba.jit(
    numba.float64[:](numba.float64[:], numba.float64[:], numba.complex128[:, :], numba.float64[:], numba.float64[:]),
    nopython=True,
    parallel=True,
    cache=True,
)
def Refl_nvary2NB(theta, lamda, n, d, sigma):
    layers = d.shape[0]
    angles = theta.shape[0]

    R = empty_like(theta)

    pre2 = pi / 180.0

    for ai in numba.prange(angles):
        n0 = n[-1, ai]
        ki = 2.0 * pi / lamda[ai]
        pre1 = 2.0 * n0 * ki

        Qi = pre1 * cmath.sqrt((n[0, ai] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)

        Qj = pre1 * cmath.sqrt((n[1, ai] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
        # Fresnel reflectivity for the interfaces
        rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[0] ** 2)
        Aj = rpj
        Qi = Qj

        for lj in range(2, layers):
            Qj = pre1 * cmath.sqrt((n[lj, ai] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
            # Fresnel reflectivity for the interfaces
            rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[lj - 1] ** 2)

            pj = cmath.exp(1.0j * d[lj - 1] * Qi)
            part = Aj * pj
            Aj = (rpj + part) / (1.0 + part * rpj)

            Qi = Qj
        R[ai] = abs(Aj) ** 2
    return R


@numba.jit(
    numba.complex128[:](numba.float64[:], numba.float64[:], numba.complex128[:, :], numba.float64[:], numba.float64[:]),
    nopython=True,
    parallel=True,
    cache=True,
)
def Amp_nvary2NB(theta, lamda, n, d, sigma):
    layers = d.shape[0]
    angles = theta.shape[0]

    A = empty(theta.shape, dtype=complex128)

    pre2 = pi / 180.0

    for ai in numba.prange(angles):
        n0 = n[-1, ai]
        ki = 2.0 * pi / lamda[ai]
        pre1 = 2.0 * n0 * ki

        Qi = pre1 * cmath.sqrt((n[0, ai] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)

        Qj = pre1 * cmath.sqrt((n[1, ai] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
        # Fresnel reflectivity for the interfaces
        rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[0] ** 2)
        Aj = rpj
        Qi = Qj

        for lj in range(2, layers):
            Qj = pre1 * cmath.sqrt((n[lj, ai] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
            # Fresnel reflectivity for the interfaces
            rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[lj - 1] ** 2)

            pj = cmath.exp(1.0j * d[lj - 1] * Qi)
            part = Aj * pj
            Aj = (rpj + part) / (1.0 + part * rpj)

            Qi = Qj
        A[ai] = Aj
    return A


def Refl_nvary2(theta, lamda, n, d, sigma, return_int=True):
    if return_int:
        return Refl_nvary2NB(theta, lamda, n, d, sigma)
    else:
        return Amp_nvary2NB(theta, lamda, n, d, sigma)



from numpy import append, arange, array, c_, cos, exp, log10, newaxis, ones, pi, r_, rollaxis, sin, sqrt

from genx.core.custom_logging import iprint


def IntElfield(theta, lamda, n, d):
    # Length of k-vector in vaccum
    k = 2 * pi / lamda
    # Calculates the wavevector in each layer
    Qj = 2 * k * sqrt(n[:, newaxis] ** 2 - cos(theta * pi / 180) ** 2)
    # print Qj
    # Fresnel reflectivity for the interfaces
    rp = (Qj[1:] - Qj[:-1]) / (Qj[1:] + Qj[:-1])
    tp = 1 + rp
    # tp=2*Qj[1:]/(Qj[1:]+Qj[:-1])
    # print tp.shape
    # Beta=k*d*n[:-1]*sin(theta*pi/180)
    # print Beta.shape
    D = d[:, newaxis] * ones(theta.shape)
    # E_p=[1]
    # E_m=[0]
    E_p = [1 * ones(theta.shape)]
    E_m = [0 * ones(theta.shape)]
    # print E_p[0].shape
    # print D[0].shape
    # print Qj[0].shape
    # for i in range(0,len(rp)):
    #    E_p.append(1/tp[i]*(E_p[i]*exp(-1.0J*D[i]*Qj[i]/2)+E_m[i]*rp[i]*exp(1.0J*D[i]*Qj[i]/2)))
    #    E_m.append(1/tp[i]*(E_m[i]*exp(1.0J*D[i]*Qj[i]/2)+E_p[i]*rp[i]*exp(-1.0J*D[i]*Qj[i]/2)))
    for i in range(0, len(rp)):
        E_p.append(1 / tp[i] * (E_p[i] * exp(-1.0j * D[i] * Qj[i] / 2) + E_m[i] * rp[i] * exp(1.0j * D[i] * Qj[i] / 2)))
        E_m.append(1 / tp[i] * (E_m[i] * exp(1.0j * D[i] * Qj[i] / 2) + E_p[i] * rp[i] * exp(-1.0j * D[i] * Qj[i] / 2)))
    # print E_p
    E_p = array(E_p[1:])
    E_m = array(E_m[1:])
    return E_p / E_p[-1], E_m / E_p[-1], Qj / 2


def AmpElfield_q(k, kx, lamda, n, d):
    # Length of k-vector in vaccum
    k = 2 * pi / lamda
    # Calculates the wavevector in each layer
    # Qj=2*k*sqrt(n[:,newaxis]**2-cos(theta*pi/180)**2)
    Qj = 2 * sqrt(k**2 * n[:, newaxis] ** 2 - kx**2)
    # Fresnel reflectivity for the interfaces
    rp = (Qj[1:] - Qj[:-1]) / (Qj[1:] + Qj[:-1])
    # print rp
    tp = 1 + rp
    # tp=2*Qj[1:]/(Qj[1:]+Qj[:-1])
    # print tp.shape
    # Beta=k*d*n[:-1]*sin(theta*pi/180)
    # print Beta.shape
    D = d[:, newaxis] * ones(kx.shape)
    # print D.shape
    # print rp.shape
    # E_p=[1]
    # E_m=[0]
    # old version
    E_p = [1 * ones(kx.shape)]
    E_m = [0 * ones(kx.shape)]
    for i in range(0, len(rp)):
        E_p.append(1 / tp[i] * (E_p[i] * exp(-1.0j * D[i] * Qj[i] / 2) + E_m[i] * rp[i] * exp(1.0j * D[i] * Qj[i] / 2)))
        E_m.append(1 / tp[i] * (E_m[i] * exp(1.0j * D[i] * Qj[i] / 2) + E_p[i] * rp[i] * exp(-1.0j * D[i] * Qj[i] / 2)))
    # Added 020905 for testing another T amplitude (The Transmitted amp. trough the interface)
    # print E_p
    # End old version
    E_p = array(E_p)
    E_m = array(E_m)
    # print E_p.shape
    # print Qj.shape
    dtemp = r_[d, 0]
    # print dtemp
    dtemp = dtemp[:, newaxis] * ones(kx.shape)
    E_p = E_p * exp(-1.0j * dtemp * Qj / 2)
    E_m = E_m * exp(1.0j * dtemp * Qj / 2)
    E_p = list(E_p)
    E_m = list(E_m)
    # End addition
    E_p.reverse()
    E_m.reverse()
    Qj = list(Qj)
    Qj.reverse()
    Qj = array(Qj)
    # Removed 020905 as said above and replaced
    E_p = array(E_p)  # Old version
    # E_p=array(E_p)
    # End addition
    E_m = array(E_m)
    q = Qj / 2  # Old ver. replaced 020905 with:
    # return (E_p[:-1]/E_p[0],E_m[:-1]/E_p[0],q[:-1]) # Last correspond to kz
    # testar amplitud i lager j+1 istallet
    return E_p[1:] / E_p[0], E_m[1:] / E_p[0], q[1:]  # Last correspond to kz


def AmpElfield2(kx, k, n, z):
    kz = sqrt(n[:, newaxis] ** 2 * k**2 - kx**2)
    r = (kz[:-1] - kz[1:]) / (kz[:-1] + kz[1:])
    t = 1 + r
    X = 0 * ones(kx.shape)
    for i in range(len(n) - 2, -1, -1):
        X = (
            exp(-2j * kz[i] * z[i])
            * (r[i] + X * exp(2j * kz[i + 1] * z[i]))
            / (1 + r[i] * X * exp(2j * kz[i + 1] * z[i]))
        )
        iprint(i)
    # X=reflected amplitude...
    r = (kz[1:] - kz[:-1]) / (kz[:-1] + kz[1:])
    t = 1 + r
    R = [X]
    T = [1 * ones(X.shape)]
    for i in range(0, len(n) - 1, 1):
        R.append(
            1
            / t[i]
            * (T[i] * r[i] * exp(-1.0j * (kz[i + 1] + kz[i]) * z[i]) + R[i] * exp(-1.0j * (kz[i + 1] - kz[i]) * z[i]))
        )
        T.append(
            1
            / t[i]
            * (T[i] * exp(1.0j * (kz[i + 1] - kz[i]) * z[i]) + R[i] * r[i] * exp(1.0j * (kz[i + 1] + kz[i]) * z[i]))
        )
    # R[len(n)-1]=0*ones(X.shape)
    R = array(R)
    T = array(T)
    return T, R, kz


from .xrmr import dot2, inv2


def AmpElfield_test(th, lam, n, d, dz=0.1):
    """Added 20110317 to see if I can do calcs of the fields inside a multilayer"""
    kz = 2 * pi / lam * sin(th * pi / 180)
    k = 2 * pi / lam
    # n = r_[n, n[-1]]
    # d = r_[d, d[-1]]
    kz = sqrt((n[:, newaxis] ** 2 - 1) * k**2 + kz**2)
    # print kz.shape, d.shape
    a = exp(-1.0j * kz * d[:, newaxis])[1:, :]
    # print a.shape
    # print kz[newaxis, newaxis, :, :].shape
    r = (kz[:-1] - kz[1:]) / (kz[:-1] + kz[1:])
    M = (
        1
        / kz[newaxis, newaxis, :-1, :]
        / 2
        * array([[(kz[1:] + kz[:-1]) * a, (kz[:-1] - kz[1:]) / a], [(kz[:-1] - kz[1:]) * a, (kz[1:] + kz[:-1]) / a]])
    )
    # M = array([[a, r*a],[r/a, 1/a]])
    M = rollaxis(M, 2, 0)[::, :, :, :]
    # "quick" version
    # MM = reduce(dot2, M)
    MMcum = [M[0]]
    for i in range(1, M.shape[0]):
        MMcum.append(dot2(MMcum[-1], M[i]))
    MM = MMcum[-1]

    R0 = MM[1, 0] / MM[0, 0]
    TN = 1 / M[0, 0]
    R = abs(R0) ** 2
    T = abs(TN) ** 2

    # Calc el. fileds as a function of z
    Rlay = R0
    Tlay = 1.0
    z = array([0])
    iprint("R0 ", R0.shape)
    E = array(abs(R0 + 1.0))[:, newaxis] * ones((th.shape[0], 1))
    # E = array([])
    iprint(kz.shape)
    iprint((th.shape[0], 1))
    iprint(ones((th.shape[0], 1)).shape)
    iprint(E.shape)
    for (
        di,
        kzi,
        M,
    ) in zip(d[1:], kz[1:], MMcum):
        Mlay = inv2(M)
        Rlay_new = Mlay[0, 0] * Tlay + Mlay[0, 1] * Rlay
        Tlay_new = Mlay[1, 0] * Tlay + Mlay[1, 1] * Rlay
        znew = arange(0, di, dz)
        z = append(z, znew + z[-1])
        # print kzi.shape
        a = exp(-1.0j * kzi[:, newaxis] * (di - znew))
        # print (abs(Rlay_new[:, newaxis]*a + Tlay_new[:, newaxis]/a)**2).shape
        E = c_[E, abs(Rlay_new[:, newaxis] * a + Tlay_new[:, newaxis] / a)]
        Rlay = Rlay_new
        Tlay = Tlay_new

    # print MM.shape
    iprint(E.shape)
    # E = E.reshape((th.shape[0], z.shape[0]))
    return T, R, z, E


if __name__ == "__main__":
    # R=[]
    # for x in arange(0,2,0.01):
    #    (E_p,E_m)=IntElfield(x,1.54,array([1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1]),array([0,80,20,80,20,80,20]))
    #    R.append(abs(E_m[-1]/E_p[-1])**2)
    import pylab

    theta = arange(0, 1 + 0.001666670, 0.005)
    # (E_p,E_m,Qj)=IntElfield(theta,1.54,array([1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1]),array([0,80,20,80,20,80,20]))
    # (E_p,E_m,Qj)=IntElfield(theta,1.540562,array([1-.15E-04/2+0.35E-06j/2,1-.45E-04/2+0.60E-05j/2,1]),array([0,1000]))
    from .paratt import Refl

    # c=Refl(theta,1.54,array([1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1]),array([80,20,80,20,80,20]),0)
    c = Refl(
        theta, 1.54, array([1 - 7.57e-6 + 1.73e-7j, 1 - 2.24e-5 + 2.89e-6j, 1]), array([0, 1000, 0]), array([0, 0, 0])
    )
    # gplt.plot(theta,log10(abs(E_m[-1]/E_p[-1])**2),theta,log10(abs(c**2)))
    # with open('Specrefl.dat','r') as f:
    #     t=io.read_array(f)
    #     t=transpose(t)
    # gplt.plot(theta,log10(abs(E_m[-1]/E_p[-1])**2),t[0],log10(t[1]))
    k = 2 * pi / 1.54
    n = [1, 1 - 2.24e-5 + 2.89e-6j, 1 - 7.57e-6 + 1.73e-7j]
    n = array(n)
    z = array([0, -1000])
    # (T,R,k)=AmpElfield2(k*cos(theta*pi/180),k*ones(theta.shape), n,z)
    # (T,R,k)=IntElfield(theta,1.54, n, array([0,1000]))
    (T, R, zc, E) = AmpElfield_test(theta, 1.54, n, array([0, 1000, 0]))
    # print T[0]
    pylab.subplot(211)
    pylab.plot(theta, log10(abs(R)), theta, log10(abs(c)), ".-")
    pylab.ylabel("R")
    pylab.xlabel("theta [deg]")
    pylab.legend(("Abeles", "Parratt"))
    pylab.subplot(212)
    pylab.contourf(theta, zc, E.transpose())
    # print zc.shape, zc[-1]
    # print log10(E[0,:]).shape
    # pylab.plot(zc, (E[1,:]))
    pylab.show()



# Functions for instrumental resolution corrections and area of illumination
# these are to be included in the Models class and Refl class
# Note that these function assumes a Gaussian spread in both the angular as well
# as in real space

# THis one really needs scipy
from numpy import (abs, allclose, append, arange, arccos, arcsin, arctan, arctan2, array, bitwise_not, c_, complex64,
                   complex128, conj, convolve, cos, cumsum, dot, exp, float32, float64, imag, isscalar, linspace,
                   logical_not, maximum, minimum, ndarray, newaxis, not_equal, ones, ones_like, pi, r_, real, rollaxis,
                   sin, sqrt, sum, tan, transpose, where, zeros, zeros_like)

try:
    # trapz removed in numpy 2.x
    from numpy import trapezoid as trapz
except ImportError:
    from numpy import trapz

from scipy.special import erf

from genx.core.custom_logging import iprint

rad = pi / 180.0
sqrt2 = sqrt(2.0)


# Intensity correction Gaussian beamprofile
def GaussIntensity(alpha, s1, s2, sigma_x):
    sinalpha = sin(alpha * rad)
    if s1 == s2:
        return erf(s2 / sqrt2 / sigma_x * sinalpha)
    else:
        common = sinalpha / sqrt2 / sigma_x
        return (erf(s2 * common) + erf(s1 * common)) / 2.0


# Specular foorprintcorrections square beamprofile
def SquareIntensity(alpha, slen, beamwidth):
    F = slen / beamwidth * sin(alpha * rad)
    return where(F <= 1.0, F, ones(F.shape))


####################################################################
## Resolution Functions (Normal Distributed=Gaussian)
#####################################################################

# Full Convlutions - vayring resolution


# Function to create a 1D vector for the resolution with the
# positions to calculate the reflectivity Qret and the weight
# of each point weight
# Inputs: Q - the Q values
#         dQ - the resolution
#         points - the number of points for the convolution
#         range how far the gaussian should be convoluted
def ResolutionVector(Q, dQ, points, range=3):
    Qres = Q + dQ * linspace(-range, range, points)[:, newaxis]
    weight = 1 / sqrt(2 * pi) / dQ * exp(-((transpose(Q[:, newaxis]) - Qres) ** 2) / dQ**2 / 2)
    Qret = Qres.flatten()
    return Qret, weight


# Include the resolution with Qret and weight calculated from ResolutionVector
# and I the calculated intensity at each point. returns the intensity
def ConvoluteResolutionVector(Qret, I, weight):
    Qret2 = Qret.reshape(weight.shape[0], weight.shape[1])
    I2 = I.reshape(weight.shape[0], weight.shape[1])
    norm_fact = trapz(weight, x=Qret2, axis=0)
    Int = trapz(I2 * weight, x=Qret2, axis=0) / norm_fact
    return Int


# Fast convlution - constant resolution
# constant spacing between data!
def ConvoluteFast(Q, I, dQ, range=3):
    Qstep = Q[1] - Q[0]
    resvector = arange(-range * dQ, range * dQ + Qstep, Qstep)
    weight = 1 / sqrt(2 * pi) / dQ * exp(-(resvector**2) / dQ**2 / 2)
    Iconv = convolve(
        r_[ones(resvector.shape) * I[0], I, ones(resvector.shape) * I[-1]], weight / weight.sum(), mode="same"
    )[resvector.shape[0] : -resvector.shape[0]]
    return Iconv


# Fast convolution - varying resolution
# constant spacing between the dat.
def ConvoluteFastVar(Q, I, dQ, range=3):
    Qstep = Q[1] - Q[0]
    steps = max(dQ * ones(Q.shape)) * range / Qstep
    weight = 1 / sqrt(2 * pi) / dQ * exp(-((Q[:, newaxis] - Q) ** 2) / dQ**2 / 2)
    Itemp = I[:, newaxis] * ones(I.shape)
    norm_fact = trapz(weight, axis=0)
    Int = trapz(Itemp * weight, axis=0) / norm_fact
    return Int


def QtoTheta(wavelength, Q):
    return arcsin(wavelength / 4.0 / pi * Q) / rad


def TwoThetatoQ(wavelength, TwoTheta):
    return 4.0 * pi / wavelength * sin(rad / 2.0 * TwoTheta)


########## Neutron Polarization #################
def get_pol_matrix(p1, p2, F1, F2):
    # @formatter:off
    P1 = array(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [F1, 0, (1.0 - F1), 0],
            [0, F1, 0, (1.0 - F1)],
        ]
    )
    P2 = array(
        [
            [1, 0, 0, 0],
            [F2, (1.0 - F2), 0, 0],
            [0, 0, 1, 0],
            [0, 0, F2, (1.0 - F2)],
        ]
    )
    P3 = array(
        [
            [(1.0 - p1), 0, p1, 0],
            [0, (1.0 - p1), 0, p1],
            [p1, 0, (1.0 - p1), 0],
            [0, p1, 0, (1.0 - p1)],
        ]
    )
    P4 = array(
        [
            [(1.0 - p2), p2, 0, 0],
            [p2, (1.0 - p2), 0, 0],
            [0, 0, (1.0 - p2), p2],
            [0, 0, p2, (1.0 - p2)],
        ]
    )
    # @formatter:on
    P = dot(dot(P1, P2), dot(P3, P4))
    return P


from . import USE_NUMBA

if USE_NUMBA:
    # try to use numba to speed up the calculation intensive functions:
    try:
        from .instrument_numba import GaussIntensity, QtoTheta, ResolutionVector, SquareIntensity, TwoThetatoQ
    except Exception as e:
        iprint("Could not use numba, no speed up from JIT compiler:\n" + str(e))



"""
==============================
:mod:`gsecars_ctr` data loader
==============================

Loads the data from whitespace seperated column formatted ascii data files.
It is intended for surface x-ray diffraction data where the data sets consists
of rod scans along the l-direction (perpendicular to the surface). The plugin
sorts each rod with equal h and k values into one data sets. The l-direction 
is also sorted.

The default columns are the following:
* 1st column h values; 2nd column k values; 3rd values l values;
* 4th column Intensites; 5th column The standard deviation of the intensities;
* 6th column L of first Bragg peak, 7th column L spacing of Bragg peaks
 The other settings are just as in the default data loader.

The h,k values is stored as extra data in data.extra_data dictonary as h and k.

If the file contains a first headerline with column titles it is used
to populate the extra_data with these columns automatically.
"""

import numpy as np

from ..data_loader_framework import Template
from ..utils import ShowWarningDialog

try:
    import wx

    from wx.lib.masked import NumCtrl
except ImportError:

    class void:
        pass

    wx = void()
    wx.Dialog = void


class Plugin(Template):
    REQUIRED_COLUMNS = ["h", "k", "l", "f", "ferr"]

    def __init__(self, parent):
        Template.__init__(self, parent)
        self.h_col = 0
        self.k_col = 1
        self.l_col = 2
        self.I_col = 3
        self.eI_col = 4
        self.LB_col = 5
        self.dL_col = 6
        self.comment = "#"
        self.skip_rows = 0
        self.delimiter = None

    def LoadData(self, dataset, filename, data_id=0):
        """LoadData(self, dataset, filename) --> none

        Loads the data from filename into the dataset.
        """
        try:
            load_array = np.loadtxt(filename, delimiter=self.delimiter, comments=self.comment, skiprows=self.skip_rows)
        except Exception as e:
            ShowWarningDialog(
                self.parent,
                "Could not load the file: "
                + filename
                + " \nPlease check the format.\n\n numpy.loadtxt"
                + " gave the following error:\n"
                + str(e),
            )
        else:
            l1 = open(filename, "r").readline().strip()
            if l1.startswith("#"):
                header_cols = l1.strip("#").strip().split()
                hc_lower = [c.lower() for c in header_cols]
                if all([c in hc_lower for c in self.REQUIRED_COLUMNS]):
                    # all necessary columns in header line, overwrite default columns
                    self.h_col = hc_lower.index("h")
                    self.k_col = hc_lower.index("k")
                    self.l_col = hc_lower.index("l")
                    self.I_col = hc_lower.index("f")
                    self.eI_col = hc_lower.index("ferr")
                    self.LB_col = 0
                    self.dL_col = 0
                    extra_columns = {}
                    for i, col in enumerate(header_cols):
                        if col.lower() in self.REQUIRED_COLUMNS:
                            continue
                        extra_columns[col] = i
            else:
                extra_columns = {"LB": self.LB_col, "dL": self.dL_col}
            # For the freak case of only one data point
            if len(load_array.shape) < 2:
                load_array = np.array([load_array])
            # Check so we have enough columns
            if load_array.shape[1] - 1 < max(
                self.h_col, self.k_col, self.l_col, self.I_col, self.eI_col, self.LB_col, self.dL_col
            ):
                ShowWarningDialog(
                    self.parent,
                    "The data file does not contain"
                    + "enough number of columns. It has "
                    + str(load_array.shape[1])
                    + " columns. Rember that the column index start at zero!",
                )
                # Okay now we have showed a dialog lets bail out ...
                return
            # The data is set by the default Template.__init__ function, neat hu
            # Note that the loaded data goes into *_raw so that they are not
            # changed by the transforms

            # Create an record array so we can sort the data properly
            data = np.rec.fromarrays(
                [
                    load_array[:, self.h_col].round().astype(type(1)),
                    load_array[:, self.k_col].round().astype(type(1)),
                    load_array[:, self.l_col],
                    load_array[:, self.I_col],
                    load_array[:, self.eI_col],
                ]
                + [load_array[:, idx] for idx in extra_columns.values()],
                names=",".join(["h", "k", "l", "I", "eI"] + list(extra_columns.keys())),
            )
            # Sort the data
            data.sort(order=("h", "k", "l"))
            i = 0
            while i < len(data):
                # Find all the data for each rod
                tmp = data.compress(np.bitwise_and(data["h"] == data[i]["h"], data["k"] == data[i]["k"]))
                self.data.add_new("(%i, %i)" % (tmp["h"][0], tmp["k"][0]))
                self.data[-1].x_raw = tmp["l"]
                self.data[-1].y_raw = tmp["I"]
                self.data[-1].error_raw = tmp["eI"]
                # Run the commands on the data - this also sets the x,y, error memebers
                # of that data item.
                self.data[-1].run_command()

                # add all extra columns
                self.data[-1].set_extra_data("h", tmp["h"], "h")
                self.data[-1].set_extra_data("k", tmp["k"], "k")
                for key in extra_columns.keys():
                    self.data[-1].set_extra_data(key, tmp[key], key)
                # Increase the index
                i += len(tmp)

    def SettingsDialog(self):
        """SettingsDialog(self) --> None

        This function should - if necessary implement a dialog box
        that allows the user set import settings for example.
        """
        col_values = {
            "I": self.I_col,
            "h": self.h_col,
            "k": self.k_col,
            "l": self.l_col,
            "I error": self.eI_col,
            "LB": self.LB_col,
            "dL": self.dL_col,
        }
        misc_values = {"Comment": str(self.comment), "Skip rows": self.skip_rows, "Delimiter": str(self.delimiter)}
        dlg = SettingsDialog(self.parent, col_values, misc_values)
        if dlg.ShowModal() == wx.ID_OK:
            col_values = dlg.GetColumnValues()
            misc_values = dlg.GetMiscValues()
            self.h_col = col_values["h"]
            self.k_col = col_values["k"]
            self.l_col = col_values["l"]
            self.I_col = col_values["I"]
            self.eI_col = col_values["I error"]
            self.LB_col = col_values["LB"]
            self.dL_col = col_values["dL"]
            self.comment = misc_values["Comment"]
            self.skip_rows = misc_values["Skip rows"]
            self.delimiter = misc_values["Delimiter"]
            self.SetStatusText("Changed import settings")
        else:
            self.SetStatusText("No change to import settings")
        dlg.Destroy()


class SettingsDialog(wx.Dialog):

    def __init__(self, parent, col_values, misc_values):
        wx.Dialog.__init__(self, parent, -1, "Data loader settings")

        box_sizer = wx.BoxSizer(wx.HORIZONTAL)

        # Make the box for putting in the columns
        col_box = wx.StaticBox(self, -1, "Columns")
        col_box_sizer = wx.StaticBoxSizer(col_box, wx.VERTICAL)

        # col_values = {'y': 1,'x': 0,'y error': 1}
        col_grid = wx.GridBagSizer(len(col_values), 2)
        self.col_controls = col_values.copy()
        keys = list(col_values.keys())
        keys.sort()
        for i, name in enumerate(keys):
            text = wx.StaticText(self, -1, name + ": ")
            control = wx.SpinCtrl(self)
            control.SetRange(0, 100)
            control.SetValue(col_values[name])
            col_grid.Add(text, (i, 0), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
            col_grid.Add(control, (i, 1), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
            self.col_controls[name] = control

        col_box_sizer.Add(col_grid, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        box_sizer.Add(col_box_sizer, 0, wx.ALL | wx.EXPAND, 5)

        col_box = wx.StaticBox(self, -1, "Misc")
        col_box_sizer = wx.StaticBoxSizer(col_box, wx.VERTICAL)

        # Lets add another box for comments and rows to skip
        # misc_values = {'Comment': '#', 'Skip rows': 0,'Delimiter': 'None'}
        col_grid = wx.GridBagSizer(len(misc_values), 2)
        self.misc_controls = misc_values.copy()
        keys = list(misc_values.keys())
        keys.sort()
        for i, name in enumerate(keys):
            text = wx.StaticText(self, -1, name + ": ")
            if type(misc_values[name]) == type(1):
                control = wx.SpinCtrl(self)
                control.SetRange(0, 100)
                control.SetValue(misc_values[name])
            else:
                control = wx.TextCtrl(self, value=misc_values[name], style=wx.EXPAND)
            col_grid.Add(text, (i, 0), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
            col_grid.Add(control, (i, 1), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
            self.misc_controls[name] = control

        col_box_sizer.Add(col_grid, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        box_sizer.Add(col_box_sizer, 0, wx.ALL | wx.EXPAND, 5)

        button_sizer = wx.StdDialogButtonSizer()
        okay_button = wx.Button(self, wx.ID_OK)
        okay_button.SetDefault()
        button_sizer.AddButton(okay_button)
        button_sizer.AddButton(wx.Button(self, wx.ID_CANCEL))
        button_sizer.Realize()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(box_sizer, 1, wx.GROW, 20)
        line = wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW, 30)

        sizer.Add(button_sizer, 0, flag=wx.ALIGN_RIGHT, border=20)
        self.SetSizer(sizer)

        sizer.Fit(self)
        self.Layout()

    def GetColumnValues(self):
        values = {}
        for key in self.col_controls:
            values[key] = self.col_controls[key].GetValue()
        return values

    def GetMiscValues(self):
        values = {}
        for key in self.misc_controls:
            val = self.misc_controls[key].GetValue()
            if type(val) == type("") or type(val) == type(""):
                if val.lower() == "none":
                    val = None
            values[key] = val
        return values



import math

import numba

from numpy import empty, empty_like, ndarray

pi = math.pi
rad = pi / 180.0
sqrt2 = math.sqrt(2.0)


def GaussIntensity(alpha, s1, s2, sigma_x):
    if type(s1) is ndarray:
        return GaussIntensityVector(alpha, s1, s2, sigma_x)
    else:
        return GaussIntensityScalar(alpha, s1, s2, sigma_x)


@numba.jit(
    numba.float64[:](numba.float64[:], numba.float64, numba.float64, numba.float64),
    nopython=True,
    parallel=True,
    cache=True,
)
def GaussIntensityScalar(alpha, s1, s2, sigma_x):
    I = empty_like(alpha)
    if s1 == s2:
        for ai in numba.prange(alpha.shape[0]):
            sinalpha = math.sin(alpha[ai] * rad)
            I[ai] = math.erf(s2 / sqrt2 / sigma_x * sinalpha)
    else:
        for ai in numba.prange(alpha.shape[0]):
            sinalpha = math.sin(alpha[ai] * rad)
            I[ai] = (math.erf(s1 / sqrt2 / sigma_x * sinalpha) + math.erf(s2 / sqrt2 / sigma_x * sinalpha)) / 2.0
    return I


@numba.jit(
    numba.float64[:](numba.float64[:], numba.float64[:], numba.float64[:], numba.float64),
    nopython=True,
    parallel=True,
    cache=True,
)
def GaussIntensityVector(alpha, s1, s2, sigma_x):
    I = empty_like(alpha)
    for ai in numba.prange(alpha.shape[0]):
        sinalpha = math.sin(alpha[ai] * rad)
        I[ai] = (math.erf(s1[ai] / sqrt2 / sigma_x * sinalpha) + math.erf(s2[ai] / sqrt2 / sigma_x * sinalpha)) / 2.0
    return I


@numba.jit(numba.float64[:](numba.float64[:], numba.float64, numba.float64), nopython=True, parallel=True, cache=True)
def SquareIntensity(alpha, slen, beamwidth):
    I = empty_like(alpha)
    scale = slen / beamwidth
    for ai in numba.prange(alpha.shape[0]):
        F = scale * math.sin(alpha[ai] * rad)
        I[ai] = min(1.0, F)
    return I


@numba.jit(numba.float64[:](numba.float64, numba.float64[:]), nopython=True, parallel=True, cache=True)
def TwoThetatoQScalar(wavelength, TwoTheta):
    Q = empty_like(TwoTheta)
    scale = 4.0 * pi / wavelength
    thscale = rad / 2.0
    for li in numba.prange(Q.shape[0]):
        Q[li] = math.sin(TwoTheta[li] * thscale) * scale
    return Q


@numba.jit(numba.float64[:](numba.float64[:], numba.float64[:]), nopython=True, parallel=True, cache=True)
def TwoThetatoQVector(wavelength, TwoTheta):
    Q = empty_like(TwoTheta)
    scale = 4.0 * pi
    thscale = rad / 2.0
    for li in numba.prange(Q.shape[0]):
        Q[li] = math.sin(TwoTheta[li] * thscale) * scale / wavelength[li]
    return Q


def TwoThetatoQ(wavelength, TwoTheta):
    if type(wavelength) is ndarray:
        return TwoThetatoQVector(wavelength, TwoTheta)
    else:
        return TwoThetatoQScalar(wavelength, TwoTheta)


@numba.jit(numba.float64[:](numba.float64, numba.float64[:]), nopython=True, parallel=True, cache=True)
def QtoThetaScalar(wavelength, Q):
    Theta = empty_like(Q)
    scale = wavelength / 4.0 / pi
    for li in numba.prange(Q.shape[0]):
        Theta[li] = math.asin(scale * Q[li]) / rad
    return Theta


@numba.jit(numba.float64[:](numba.float64[:], numba.float64[:]), nopython=True, parallel=True, cache=True)
def QtoThetaVector(wavelength, Q):
    Theta = empty_like(Q)
    scale = 1.0 / 4.0 / pi
    for li in numba.prange(Q.shape[0]):
        Theta[li] = math.asin(scale * wavelength[li] * Q[li]) / rad
    return Theta


def QtoTheta(wavelength, Q):
    if type(wavelength) is ndarray:
        return QtoThetaVector(wavelength, Q)
    else:
        return QtoThetaScalar(wavelength, Q)


#
@numba.jit(
    numba.types.Tuple((numba.float64[:], numba.float64[:, :]))(
        numba.float64[:], numba.float64, numba.int16, numba.float64
    ),
    nopython=True,
    parallel=True,
    cache=True,
)
def ResolutionVectorScalar(Q, dQ, points, range):
    Qres = empty(Q.shape[0] * points)
    weight = empty((points, Q.shape[0]))
    Qstep = 2.0 * range / (points - 1) * dQ
    NQ = Q.shape[0]

    wscale = 1.0 / math.sqrt(2.0 * pi) / dQ

    for ri in numba.prange(points):
        dq = (ri - (points - 1.0) / 2.0) * Qstep
        for qi in numba.prange(NQ):
            Qi = Q[qi]
            Qj = Qi + dq
            Qres[qi + ri * NQ] = Qj
            weight[ri, qi] = wscale * math.exp(-((Qi - Qj) ** 2) / (dQ) ** 2 / 2.0)

    return (Qres, weight)


@numba.jit(
    numba.types.Tuple((numba.float64[:], numba.float64[:, :]))(
        numba.float64[:], numba.float64[:], numba.int16, numba.float64
    ),
    nopython=True,
    parallel=True,
    cache=True,
)
def ResolutionVectorVector(Q, dQ, points, range):
    Qres = empty(Q.shape[0] * points)
    weight = empty((points, Q.shape[0]))
    Qstep_scale = 2.0 * range / (points - 1.0)
    NQ = Q.shape[0]

    wscale = 1.0 / math.sqrt(2.0 * pi)

    for ri in numba.prange(points):
        dq_scale = (ri - (points - 1.0) / 2.0) * Qstep_scale
        for qi in numba.prange(NQ):
            dq = dq_scale * dQ[qi]
            Qi = Q[qi]
            Qj = Qi + dq
            Qres[qi + ri * NQ] = Qj
            weight[ri, qi] = wscale * math.exp(-((Qi - Qj) ** 2) / (dQ[qi]) ** 2 / 2.0) / dQ[qi]

    return (Qres, weight)


def ResolutionVector(Q, dQ, points, range=3):
    if type(dQ) is ndarray:
        return ResolutionVectorVector(Q, dQ, points, range)
    else:
        return ResolutionVectorScalar(Q, dQ, points, range)



r"""
.. module:: island_lattice.py
   :synopsis: Models to calcualte the reflectivity from a 2D island array

This module provide functions to calculate the reflectivity from gratings using the distorted wave Born approximation
(DWBA) and the kinematical theory.
"""

import numpy as np

from genx.core.custom_logging import iprint

from . import grating
from .materials import Material
from .parameters import ComplexArray, Float, Func, HasParameters, Int, List, Var

q = Var("q", np.array([0, 0, 0]), help="The scattering vector.")
e = Var("wl", 1.54, help="The wavelength of the radiation.")
z = Var("z", 0.0, help="The out-of-plane position in the sample.")
rep = Var("rep", 0, help="The repetition index in a Stack, starting with zero.")

validation_kwargs = {"q": np.array([0, 0, 0]), "z": 0, "rep": 0}


class Layer(HasParameters):
    validation_kwargs = validation_kwargs
    protected = True

    def __init__(self, **kwargs):
        self.d = Float(0, unit="AA", help="Layer thickness")
        self.sld = ComplexArray(
            0, unit="1/AA^2", help="The in-plane fourier transform of the scattering length density"
        )
        HasParameters.__init__(self, **kwargs)


class Stack(HasParameters):
    validation_kwargs = validation_kwargs
    protected = True

    def __init__(self, **kwargs):
        try:
            value = kwargs.pop("layers")
        except KeyError:
            value = []
        self.layers = List(Layer, value=value, help="The layers in the stack")
        self.reps = Int(1, help="Number of repetitions of the stack")
        HasParameters.__init__(self, **kwargs)

    def build_layer_list(self, parameters, kwargs):
        """Builds an array for layer parameters by looping trough the stacks and layers

        Parameters:
            parameters(list of string): name of a valid layer parameters
            kwargs(dict): Keyword argument to be passed to the parameter call
        """
        out = dict()
        [out.__setitem__(par, []) for par in parameters]
        kwargs["rep"] = -1
        for n in range(self.reps(**kwargs)):
            kwargs["rep"] += 1
            for layer in self.layers:
                # Evaluate the parameter value
                # print kwargs
                [out[par].append(layer.__getattribute__(par)(**kwargs)) for par in parameters]
                kwargs["z"] += layer.d(**kwargs)
        kwargs.pop("rep")
        return out


class Sample(HasParameters):
    validation_kwargs = validation_kwargs

    def __init__(self, **kwargs):
        try:
            value = kwargs.pop("stacks")
        except KeyError:
            value = []
        self.stacks = List(Stack, value=value, help="The stacks in the sample")
        self.substrate = Layer(d=0.0, sld=0 * FLayer())
        self.ambient = Layer(d=0.0, sld=0 * FLayer())

        self.a = Float(2 * np.pi, unit="AA", help="Unit cell size along the x-direction")
        self.b = Float(2 * np.pi, unit="AA", help="Unit cell size along the y-direction")

        # dictionary that holds the last evaluated layer parameters - used for checking the calcs
        self.layer_dic = {}

        HasParameters.__init__(self, **kwargs)

    def build_layer_list(self, parameters, kwargs):
        """Builds an array for layer parameters by looping through the stacks and layers

        Parameters:
            parameters(list of string): name of a valid layer parameter
            kwargs(dict): Keyword argument to be passed to the parameter call
        """
        kwargs["z"] = 0.0
        out = dict()
        [out.__setitem__(par, [self.substrate.__getattribute__(par)(**kwargs)]) for par in parameters]
        for stack in self.stacks:
            stack_dict = stack.build_layer_list(parameters, kwargs)
            [out[par].extend(stack_dict[par]) for par in parameters]
        kwargs["z"] = 0.0
        [out[par].append(self.ambient.__getattribute__(par)(**kwargs)) for par in parameters]
        return out

    def simulate(self, instrument, h, k, qz):
        self.layer_dic = {}
        # Add a q = 0 point for the mean scattering potential
        q = np.c_[h * 2 * np.pi / self.a(), k * 2 * np.pi / self.b(), -qz].T
        kwargs = {"z": 0.0, "q": q}
        wavelength = instrument.wl(**kwargs)
        kwargs["wl"] = wavelength

        area_uc = (self.a * self.b)(**kwargs)
        dic = self.build_layer_list(["d", "sld"], kwargs)
        vf, d = 4 * np.pi * np.array(dic["sld"]), np.array(dic["d"])
        kwargs["q"] = np.array([0, 0, 0])
        vmean = 4 * np.pi * np.array(self.build_layer_list(["sld"], kwargs)["sld"]) / area_uc
        self.layer_dic["sld_mean"] = vmean / 4 / np.pi
        self.layer_dic["d"] = d
        iprint(vmean.shape, vf.shape)
        k_in, k_out = instrument.reflectometer_3axis_kinout(q)
        R = grating.coherent_refl(k_in, k_out, wavelength, vf.T, vmean, d, area_uc, kin=False)
        return R


class Instrument(HasParameters):
    validation_kwargs = validation_kwargs

    def __init__(self, **kwargs):
        self.wl = Float(1.54, unit="AA", help="Wavelength of the used radiation")
        HasParameters.__init__(self, **kwargs)

    def reflectometer_3axis_kinout(self, q):
        """Tranforms q values to k_in and k_out vectors for a reflectometer setup with a sample rotation circle.

        Parameters:
            q(array): an array for qx, qy and qz, size (3xM)

        Returns:
            k_in (array): The incomming wavevector, size (3xM)
            k_out (array): The outgoing wavevector, size (3xM)
        """
        qx = q[0]
        qy = q[1]
        qz = q[2]
        k = 2 * np.pi / self.wl()
        qp = np.sqrt(qx * qx + qy * qy)
        q_abs = np.sqrt(qx * qx + qy * qy + qz * qz)
        phi = np.arctan2(qy, qx)
        cos_phi = np.cos(phi)
        sin_phi = np.sin(phi)

        apb = np.arcsin(q_abs / 2.0 / k)
        amb = np.arctan2(qp, -qz) % np.pi
        alpha = apb + amb
        beta = apb - amb
        cos_alpha = np.cos(alpha)
        cos_beta = np.cos(beta)
        k_in = k * np.c_[cos_alpha * cos_phi, cos_alpha * sin_phi, np.sin(alpha)].T
        k_out = k * np.c_[cos_beta * cos_phi, cos_beta * sin_phi, -np.sin(beta)].T
        return k_in, k_out


class FLayer(Func):
    """The in-plane fourier transform of an in-plane infinite layer"""

    def function(self, **kwargs):
        q = kwargs["q"]
        return np.where(np.bitwise_and(q[0] == 0, q[1] == 0), 1.0, 0.0)


class FCircle(Func):
    """The in-plane fourier transform of a circle"""

    def __init__(self, **kwargs):
        self.r = Float(1.0, help="The radius of the circle")
        Func.__init__(self, **kwargs)

    def function(self, **kwargs):
        q = kwargs["q"]
        r = self.r(**kwargs)
        q_r = np.sqrt(q[0] ** 2 + q[1] ** 2)
        return 2 * np.pi * r**2 * grating.jinc(q_r * r)


if __name__ == "__main__":
    radius = Float(1.0)
    slope = Float(1)
    mat_Fe = Material("Fe", density=7.87)
    Fe = Layer(d=rep * 0.1 + 10.0, sld=mat_Fe.sld_x * FCircle(r=radius + z * slope))
    Fe.d = 20.0
    q = np.array([0, 0, 0])
    iprint(Fe.sld(q=q))
    iprint(Fe.d)
    ML = Stack(
        reps=10,
        layers=[
            Fe,
        ],
    )

    s = Sample(
        stacks=[
            ML,
        ]
    )
    iprint(isinstance(s, HasParameters), s.__class__.__name__)
    inst = Instrument()

    iprint("D with a gradient: ", s.build_layer_list(["d"], {})["d"])

    # print s.stacks[0].layers[0]
    qz = np.arange(0, 0.1, 0.001)
    s.simulate(inst, qz * 0 + 0.0, qz * 0 + 0.01, qz)
